/****************************************************************************
**
** Copyright (C) 2001-2003 RibbonSoft. All rights reserved.
** Copyright (c) 2007 Martin Teichmann. All rights reserved.
**
** This file is part of the KAD project, based on a file
** from the QCad project.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
**********************************************************************/

#include "main.h"

#include <locale.h>
#include <klocale.h>
#include <kaboutdata.h>

#include <kapplication.h>
#include <kcmdlineargs.h>
#include <kstandarddirs.h>
#include <kurl.h>

#ifdef __APPLE__
#include <qmacstyle_mac.h>
#endif

#include <qtimer.h>

#if QT_VERSION>=0x030200
# include <qsplashscreen.h>
  QSplashScreen *splash = 0;
#endif

#ifdef RS_SCRIPTING
#include <qsproject.h>
#endif

#include "rs_fontlist.h"
#include "rs_patternlist.h"
#include "rs_scriptlist.h"
#include "rs_settings.h"
#include "rs_system.h"
#include "rs_fileio.h"
#include "rs_filtercxf.h"
#include "rs_filterdxf.h"
#include "rs_filterdxf1.h"

#include "qg_dlginitial.h"

#include "qc_applicationwindow.h"

#ifdef RS_CAM
# include "rs_filtercam.h"
# include "rs_camdialog.h"
#endif

#ifdef RS_CAM
# include "xpm/intro_camexpert.xpm"
# include "xpm/splash_camexpert.xpm"
#else
# include "xpm/intro_qcad.xpm"
# ifndef QC_CUSTOM_SPLASH
#  include "xpm/splash_qcad.xpm"
# endif
#endif

#ifndef QC_SPLASH_TXTCOL
# define QC_SPLASH_TXTCOL Qt::black
#endif

// for image mime resources from png files
extern void qInitImages_qcadactions();

#ifdef RS_SCRIPTING
	extern void qInitImages_qcadscripting();
#endif

#ifdef QC_BUILTIN_STYLE
	extern void applyBuiltinStyle();
#endif

static const KCmdLineOptions options[] =
{
	{ "debug", I18N_NOOP("Debuging"), 0 },
	{ "+[file]", I18N_NOOP("File to open"), 0 },
	KCmdLineLastOption
};

/**
 * Main. Creates Application window.
 *
 * @todo Switch back to xpm instead of png. 
 * Cleaning up #defines.
 */

class KApplication;
int main(int argc, char** argv)
{
	KAboutData about("kad", I18N_NOOP("KAD"), VERSION,
		I18N_NOOP("KDE Aided Design"),
		KAboutData::License_GPL_V2,
		I18N_NOOP(
		"Copyright (c) 2007 RibbonSoft and Martin Teichmann"));
	about.addAuthor("Andrew Mustun", I18N_NOOP("developer"),
		"info@ribbonsoft.com", 0);
	about.addAuthor("Martin Teichmann", I18N_NOOP("KDE version"),
		"martin.teichmann@lkb.ens.fr", 0);
	KCmdLineArgs::init(argc, argv, &about);
	KCmdLineArgs::addCmdLineOptions(options);
	KApplication::addCmdLineOptions();
    	KApplication app;
	QDir dir(app.applicationDirPath());
	dir.cdUp();
	KGlobal::dirs()->addPrefix(dir.absPath());
	RS_DEBUG->setLevel(RS_Debug::D_WARNING);
	RS_DEBUG->print("param 0: %s", argv[0]);

	QFileInfo prgInfo( QFile::decodeName(argv[0]) );
	dir.cd("share");
    RS_SYSTEM->init("KAD", VERSION, "qcad", dir.absPath());
	
	RS_FILEIO->registerFilter(new RS_FilterCXF());
	RS_FILEIO->registerFilter(new RS_FilterDXF());
	RS_FILEIO->registerFilter(new RS_FilterDXF1());

	// parse command line arguments that might not need a launched program:
	RS_String lang;
	RS_String langCmd;
	RS_String unit;
	
	RS_SETTINGS->beginGroup("/Defaults");
#ifndef QC_PREDEFINED_UNIT
	unit = RS_SETTINGS->readEntry("/Unit", "Invalid");
#else
	unit = RS_SETTINGS->readEntry("/Unit", QC_PREDEFINED_UNIT);
#endif
    RS_SETTINGS->endGroup();

	// show initial config dialog:
	if (unit=="Invalid") {
		RS_DEBUG->print("main: show initial config dialog..");
		QG_DlgInitial di(NULL);
		di.setText("<font size=\"+1\"><b>Welcome to " QC_APPNAME "</b></font>"
		"<br>"
		"Please choose the unit you want to use for new drawings and your "
		"preferred language.<br>"
		"You can changes these settings later in the "
		"Options Dialog of " QC_APPNAME ".");
		QPixmap pxm(intro_xpm);
		di.setPixmap(pxm);
		if (di.exec()) {
			RS_SETTINGS->beginGroup("/Defaults");
			unit = RS_SETTINGS->readEntry("/Unit", "None");
    		RS_SETTINGS->endGroup();
		}
		RS_DEBUG->print("main: show initial config dialog: OK");
	}

    RS_FONTLIST->init();
	
    RS_PATTERNLIST->init();

    RS_SCRIPTLIST->init();

	RS_DEBUG->print("main: loading translation..");
	RS_SETTINGS->beginGroup("/Appearance");
#ifdef QC_PREDEFINED_LOCALE
		lang = RS_SETTINGS->readEntry("/Language", "");
		if (lang.isEmpty()) {
			lang=QC_PREDEFINED_LOCALE;
			RS_SETTINGS->writeEntry("/Language", lang);
		}
		langCmd = RS_SETTINGS->readEntry("/LanguageCmd", "");
		if (langCmd.isEmpty()) {
			langCmd=QC_PREDEFINED_LOCALE;
			RS_SETTINGS->writeEntry("/LanguageCmd", langCmd);
		}
#else
    lang = RS_SETTINGS->readEntry("/Language", "en");
    langCmd = RS_SETTINGS->readEntry("/LanguageCmd", "en");
#endif
	RS_SETTINGS->endGroup();
	
	RS_SYSTEM->loadTranslation(lang, langCmd);
	RS_DEBUG->print("main: loading translation: OK");

#ifdef QC_BUILTIN_STYLE //js:
	RS_DEBUG->print("main: applying built in style..");
	applyBuiltinStyle();
#endif

	QC_ApplicationWindow * appWin = new QC_ApplicationWindow();
	appWin->setCaption(QC_APPNAME);
	appWin->show();
	appWin->setFocus();

	KCmdLineArgs *args = KCmdLineArgs::parsedArgs();

	if (args->isSet("debug")) RS_DEBUG->setLevel(RS_Debug::D_DEBUGGING);

	for (int i = 0; i < args->count(); i++) {
		KURL url;
		url.setPath(args->arg(i));
		appWin->slotFileOpen(url);
	}

	if (!args->count()) {
		appWin->slotFileNew();
	}
	args->clear();

	appWin->slotRunStartScript();

	return app.exec();
}
