/****************************************************************************
**
** Copyright (C) 2001-2003 RibbonSoft. All rights reserved.
** Copyright (c) 2007 Martin Teichmann. All rights reserved.
**
** This file is part of the KAD project, it was modified from
** a similar file from the QCad project.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
**********************************************************************/

#include "qc_applicationwindow.h"

#include <fstream>

#include <qaccel.h>
#include <qaction.h>
#include <qapplication.h>
#include <qdatetime.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qkeycode.h>
#include <qmessagebox.h>
#include <qpaintdevicemetrics.h>
#include <qpainter.h>
#include <kprinter.h>
#include <qtimer.h>
#include <qtextstream.h>
#include <qvbox.h>

#include <qdockwindow.h>
#include <qeventloop.h>
#include <qtextedit.h>

#include <kaccel.h>
#include <kactionclasses.h>
#include <klocale.h>
#include <kurl.h>
#include <kfiledialog.h>

#include "rs_application.h"
#include "rs_actiondrawlinefree.h"
#include "rs_actionprintpreview.h"
#include "rs_creation.h"
#include "rs_dialogfactory.h"
#include "rs_dimaligned.h"
#include "rs_dimlinear.h"
#include "rs_dimradial.h"
#include "rs_ellipse.h"
#include "rs_hatch.h"
#include "rs_image.h"
#include "rs_fileio.h"
#include "rs_insert.h"
#include "rs_text.h"
#include "rs_settings.h"
#include "rs_script.h"
#include "rs_scriptlist.h"
#include "rs_solid.h"
#include "rs_staticgraphicview.h"
#include "rs_system.h"
#include "rs_actionlibraryinsert.h"

#include "qg_cadtoolbarmain.h"
#include "qg_colorbox.h"
#include "qg_coordinatewidget.h"
#include "qg_dlgimageoptions.h"
#include "qg_filedialog.h"
#include "qg_mousewidget.h"
#include "qg_pentoolbar.h"
#include "qg_selectionwidget.h"
#include "qg_cadtoolbarmain.h"
#include "qg_dlgimageoptions.h"
#include "qg_mousewidget.h"

#include "qc_mdiwindow.h"
#include "qc_dialogfactory.h"
#include "main.h"

QC_ApplicationWindow* QC_ApplicationWindow::appWindow = NULL;

#ifndef QC_APP_ICON
# define QC_APP_ICON "qcad.png"
#endif
#ifndef QC_APP_ICON16
# define QC_APP_ICON16 "qcad16.png"
#endif

/**
 * Constructor. Initializes the app.
 */
QC_ApplicationWindow::QC_ApplicationWindow()
        : KMDI::MainWindow(0),
        QG_MainWindowInterface()
{
	RS_DEBUG->print("QC_ApplicationWindow::QC_ApplicationWindow");

	appWindow = this;
	assistant = 0;
    
	actionHandler = new QG_ActionHandler(this);

#ifdef RS_SCRIPTING
	RS_DEBUG->print("QC_ApplicationWindow::QC_ApplicationWindow: creating scripter");
    scripter = new QS_Scripter(this, this);
	RS_DEBUG->print("QC_ApplicationWindow::QC_ApplicationWindow: creating scripter: OK");
#endif

    initView();
    initToolBar();
    initActions();
    initMenuBar();
    initStatusBar();

    dialogFactory = new QC_DialogFactory(this, optionWidget);
    RS_DialogFactory::instance()->setFactoryObject(dialogFactory);

    // Disable menu and toolbar items
    emit windowsChanged(FALSE);

	
    statusBar()->message(QC_APPNAME " Ready", 2000);
    //setFocusPolicy(WheelFocus);

    connect(tabWidget(), SIGNAL(currentChanged(QWidget*)),
	this, SLOT(slotWindowActivated(QWidget*)));
}



/**
 * Destructor.
 */
QC_ApplicationWindow::~QC_ApplicationWindow() {
	// No destruction needed: program exits anyway
}



/**
 * Runs the start script if scripting is available.
 */
void QC_ApplicationWindow::slotRunStartScript() {
	slotRunScript("autostart.qs");
}



/**
 * Runs a script. The action that triggers this slot has to carry the
 * name of the script file.
 */
void QC_ApplicationWindow::slotRunScript() {
    const QObject* s = sender();
    if (s!=NULL) {
        QString script = ((QAction*)s)->text();
		slotRunScript(script);
    }
}



/**
 * Runs the script with the given name.
 */
void QC_ApplicationWindow::slotRunScript(const QString& name) {
#ifdef RS_SCRIPTING
	RS_DEBUG->print("QC_ApplicationWindow::slotRunScript");

	if (scripter==NULL) {
		RS_DEBUG->print(RS_Debug::D_WARNING, 
			"QC_ApplicationWindow::slotRunScript: "
			"scripter not initialized");
		return;
	}

    statusBar()->message(tr("Running script '%1'").arg(name), 2000);

	QStringList scriptList = RS_SYSTEM->getScriptList();
	scriptList.append(RS_SYSTEM->getHomeDir() + "/.qcad/" + name);
	
	for (QStringList::Iterator it = scriptList.begin(); it!=scriptList.end(); ++it) {
		RS_DEBUG->print("QC_ApplicationWindow::slotRunScript: "
			"checking script '%s'", (*it).latin1());
		QFileInfo fi(*it);
		if (fi.exists() && fi.fileName()==name) {
			RS_DEBUG->print("QC_ApplicationWindow::slotRunScript: running '%s'", 
				(*it).latin1());
			scripter->runScript(*it, "main");
		}
	}
#endif
}



/**
 * Called from toolbar buttons that were added by scripts to 
 * insert blocks.
 */
void QC_ApplicationWindow::slotInsertBlock() {
    const QObject* s = sender();
    if (s!=NULL) {
        QString block = ((QAction*)s)->text();
        RS_DEBUG->print("QC_ApplicationWindow::slotInsertBlock: %s", 
			block.latin1());
		slotInsertBlock(block);
    }
}



/**
 * Called to insert blocks.
 */
void QC_ApplicationWindow::slotInsertBlock(const QString& name) {
	RS_DEBUG->print("QC_ApplicationWindow::slotInsertBlock: '%s'", name.latin1());

    statusBar()->message(tr("Inserting block '%1'").arg(name), 2000);

	RS_GraphicView* graphicView = getGraphicView();
	RS_Document* document = getDocument();
	if (graphicView!=NULL && document!=NULL) {
		RS_ActionLibraryInsert* action = 
			new RS_ActionLibraryInsert(*document, *graphicView);
		action->setFile(name);
		graphicView->setCurrentAction(action);
	}
}

/**
 * Close Event. Called when the user tries to close the app.
 */
void QC_ApplicationWindow::closeEvent(QCloseEvent* /*ce*/) {
    slotFileQuit();
}



/**
 * Handles right-clicks for moving back to the last cad tool bar.
 */
void QC_ApplicationWindow::mouseReleaseEvent(QMouseEvent* e) {
    e->accept();
}

/**
 * Initializes all QActions of the application.
 */
void QC_ApplicationWindow::initActions() {
	KActionMenu *menu;
	KToggleAction *toggle;
	KAction *action;

	KStdAction::openNew(this, SLOT(slotFileNew()), actionCollection());
	KStdAction::open(this, SLOT(slotFileOpen()), actionCollection());
	KStdAction::save(this, SLOT(slotFileSave()), actionCollection());
	KStdAction::saveAs(this, SLOT(slotFileSaveAs()), actionCollection());
	recentFiles = KStdAction::openRecent(this, SLOT(slotFileOpen()),
		actionCollection());
	new KAction(i18n("Export"), 0, this,
		SLOT(slotFileExport()), actionCollection(), "export");
	KStdAction::close(this, SLOT(slotFileClose()), actionCollection());
	KStdAction::print(this, SLOT(slotFilePrint()), actionCollection());
	toggle = new KToggleAction(i18n("Print Preview"),
		"fileprintpreview", 0, this,
		0, actionCollection(), "file_print_preview");
	connect(toggle, SIGNAL(toggled(bool)), this,
		SLOT(slotFilePrintPreview(bool)));
	KStdAction::quit(this, SLOT(slotFileQuit()), actionCollection());

	KStdAction::undo(actionHandler, SLOT(slotEditUndo()),
		actionCollection());
	KStdAction::redo(actionHandler, SLOT(slotEditRedo()),
		actionCollection());
	KStdAction::cut(actionHandler, SLOT(slotEditCut()),
		actionCollection());
	KStdAction::copy(actionHandler, SLOT(slotEditCopy()),
		actionCollection());
	KStdAction::paste(actionHandler, SLOT(slotEditPaste()),
		actionCollection());
	KStdAction::preferences(this, SLOT(slotOptionsGeneral()),
		actionCollection());
	new KAction(i18n("Options for Current Drawing"), 0, actionHandler,
		SLOT(slotOptionsDrawing()), actionCollection(),
		"optionsdrawing");

	toggle = new KToggleAction(i18n("&Grid"), "viewgrid", 0, this,
		0, actionCollection(), "view_grid");
	connect(toggle, SIGNAL(toggled(bool)), this, SLOT(slotViewGrid(bool)));
	toggle = new KToggleAction(i18n("&Draft"), "viewdraft", 0, this,
		0, actionCollection(), "view_draft");
	connect(toggle, SIGNAL(toggled(bool)), this,
		SLOT(slotViewDraft(bool)));
	new KAction(i18n("Redraw"), "zoomredraw", 0, actionHandler,
		SLOT(slotZoomRedraw()), actionCollection(),
		"view_redraw");
	KStdAction::zoomIn(actionHandler, SLOT(slotZoomIn()),
		actionCollection());
	KStdAction::zoomOut(actionHandler, SLOT(slotZoomOut()),
		actionCollection());
	KStdAction::fitToPage(actionHandler, SLOT(slotZoomAuto()),
		actionCollection())->setIcon("zoomauto");
	new KAction(i18n("Previous View"), "zoomprevious", 0, actionHandler,
		SLOT(slotZoomPrevious()), actionCollection(),
		"zoom_previous");
	new KAction(i18n("Window Zoom"), "zoomwindow", 0, actionHandler,
		SLOT(slotZoomWindow()), actionCollection(),
		"zoom_window");
	new KAction(i18n("Pan Zoom"), "zoompan", 0, actionHandler,
		SLOT(slotZoomPan()), actionCollection(),
		"zoom_pan");
	toggle = KStdAction::showStatusbar(this, 0, actionCollection());
	connect(toggle, SIGNAL(toggled(bool)), this,
		SLOT(slotViewStatusBar(bool)));
// TODO    menu->insertItem(tr("Vie&ws"), createDockWindowMenu(NoToolBars));
    // Selecting actions:
	new KAction(i18n("Focus on &Command Line"), Key_Space, this,
		SLOT(slotFocusCommandLine()), actionCollection(),
		"view_commandline");

	new KAction(i18n("Deselect All"), "qg_selectnothing",
		0, actionHandler, SLOT(slotDeselectAll()),
		actionCollection(), "deselect_all");
	new KAction(i18n("Select All"), "qg_selectall",
		0, actionHandler, SLOT(slotSelectAll()),
		actionCollection(), "select_all");
	new KAction(i18n("Select Single"), "qg_selectsingle",
		0, actionHandler, SLOT(slotSelectSingle()),
		actionCollection(), "select_single");
	new KAction(i18n("Select Contour"), "qg_selectcontour",
		0, actionHandler, SLOT(slotSelectContour()),
		actionCollection(), "select_contour");
	new KAction(i18n("Select Window"), "qg_selectwindow",
		0, actionHandler, SLOT(slotSelectWindow()),
		actionCollection(), "select_window");
	new KAction(i18n("Deselect Window"), "qg_deselectwindow",
		0, actionHandler, SLOT(slotDeselectWindow()),
		actionCollection(), "deselect_window");
	new KAction(i18n("Select Invert"), "qg_selectinvert",
		0, actionHandler, SLOT(slotSelectInvert()),
		actionCollection(), "select_invert");
	new KAction(i18n("Select Intersected"), "qg_selectinters",
		0, actionHandler, SLOT(slotSelectIntersected()),
		actionCollection(), "select_intersected");
	new KAction(i18n("Deselect Intersected"), "qg_selectinters",
		0, actionHandler, SLOT(slotDeselectIntersected()),
		actionCollection(), "deselect_intersected");
	new KAction(i18n("Select Layer"), "qg_selectlayer",
		0, actionHandler, SLOT(slotSelectLayer()),
		actionCollection(), "select_layer");

	toggle = new KToggleAction(i18n("Draw Point"), "qg_points",
		0, actionHandler, SLOT(slotDrawPoint()),
		actionCollection(), "draw_point");
	toggle->setExclusiveGroup("draw");

	menu = new KActionMenu(i18n("Draw Lines"), "qg_menuline",
		actionCollection(), "draw_lines");
	menu->setDelayed(true);
	toggle = new KToggleAction(i18n("Draw Line"), 
		"qg_linesnormal", 0, actionHandler, SLOT(slotDrawLine()),
		actionCollection(), "draw_line");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Angle"),
		"qg_linesangle", 0, actionHandler, SLOT(slotDrawLineAngle()),
		actionCollection(), "draw_line_angle");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Horizontal"),
		"qg_lineshor", 0, actionHandler,
		SLOT(slotDrawLineHorizontal()),
		actionCollection(), "draw_line_horizontal");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Vertical"),
		"qg_linesver", 0, actionHandler,
		SLOT(slotDrawLineVertical()),
		actionCollection(), "draw_line_vertical");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Rectangle"),
		"qg_linesrect", 0, actionHandler,
		SLOT(slotDrawLineRectangle()),
		actionCollection(), "draw_line_rectangle");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Parallel"),
		"qg_linespara", 0, actionHandler,
		SLOT(slotDrawLineParallel()),
		actionCollection(), "draw_lines_parallel");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Lines Parallel Through"),
		"qg_linesparathrough", 0, actionHandler,
		SLOT(slotDrawLineParallelThrough()),
		actionCollection(), "draw_lines_parallel_through");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Bisector"),
		"qg_linesbisector", 0, actionHandler,
		SLOT(slotDrawLineBisector()),
		actionCollection(), "draw_lines_bisector");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line tangent 1"),
		"qg_linestan1", 0, actionHandler,
		SLOT(slotDrawLineTangent1()),
		actionCollection(), "draw_lines_tangent_1");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Tangent 2"), 
		"qg_linestan2", 0, actionHandler,
		SLOT(slotDrawLineTangent2()),
		actionCollection(), "draw_lines_tangent_2");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Orthogonal"),
		"qg_linesorthogonal", 0, actionHandler,
		SLOT(slotDrawLineOrthogonal()),
		actionCollection(), "draw_lines_orthogonal");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Relative Angle"),
		"qg_linesrelativeangle", 0, actionHandler,
		SLOT(slotDrawLineRelAngle()),
		actionCollection(), "draw_lines_relative_angle");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Line Free"), 
		"qg_linesfree", 0, actionHandler, SLOT(slotDrawLineFree()),
		actionCollection(), "draw_lines_free");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Polygon Centered"),
		"qg_linespolygon", 0, actionHandler,
		SLOT(slotDrawLinePolygon()),
		actionCollection(), "draw_lines_polygon");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Polygon Two Points"),
		"qg_linespolygon2", 0, actionHandler,
		SLOT(slotDrawLinePolygon2()),
		actionCollection(), "draw_lines_polygon_2");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);

	menu = new KActionMenu(i18n("Draw Arcs"), "qg_menuarc",
		actionCollection(), "draw_arcs");
	menu->setDelayed(true);
	toggle = new KToggleAction(i18n("Draw Arc Center, Point, Angles"),
		"qg_arcscraa", 0, actionHandler, SLOT(slotDrawArc()), 
		actionCollection(), "draw_arc");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Arc tree Points"), 
		"qg_arcs3p", 0, actionHandler, SLOT(slotDrawArc3P()),
		actionCollection(), "draw_arc_3p");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Arc Parallel"),
		"qg_arcspara", 0, actionHandler, SLOT(slotDrawArcParallel()),
		actionCollection(), "draw_arc_parallel");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);

	menu = new KActionMenu(i18n("Draw Circles"), "qg_menucircle",
		actionCollection(), "draw_circles");
	menu->setDelayed(true);
	toggle = new KToggleAction(i18n("Circle with center and point"),
		"qg_circles", 0, actionHandler, SLOT(slotDrawCircle()),
		actionCollection(), "draw_circle");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Circle with center and radius"),
		"qg_circlescr", 0, actionHandler, SLOT(slotDrawCircleCR()),
		actionCollection(), "draw_circle_cr");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n(
		"Draw Circle with two opposite points"),
		"qg_circles2p", 0, actionHandler, SLOT(slotDrawCircle2P()),
		actionCollection(), "draw_circle_2p");
	toggle->setExclusiveGroup("draw");
	toggle = new KToggleAction(i18n("Draw Circle with three points"),
		"qg_circles3p", 0, actionHandler, SLOT(slotDrawCircle3P()),
		actionCollection(), "draw_circle_3p");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Draw Concenctric Circles"),
		"qg_circlespara", 0, actionHandler,
		SLOT(slotDrawCircleParallel()),
		actionCollection(), "draw_circle_parallel");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);

	menu = new KActionMenu(i18n("Draw Ellipses"), "qg_menuellipse",
		actionCollection(), "draw_ellipses");
	menu->setDelayed(true);
	toggle = new KToggleAction(i18n("Ellipse with center and two points"),
		"qg_ellipsesaxes", 0, actionHandler,
		SLOT(slotDrawEllipseAxis()),
		actionCollection(), "draw_ellipse_axis");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n(
		"Ellipse arc with center, two points and angle"), 
		"qg_ellipsearcsaxes", 0, actionHandler, 
		SLOT(slotDrawEllipseArcAxis()),
		actionCollection(), "draw_ellipse_arc_axis");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);

	toggle = new KToggleAction(i18n("Draw Spline"), "qg_points",
		0, actionHandler, SLOT(slotDrawSpline()),
		actionCollection(), "draw_spline");
	toggle->setExclusiveGroup("draw");

	new KAction(i18n("Draw Text"),
		"qg_menutext",
		0, actionHandler, SLOT(slotDrawText()),
		actionCollection(), "draw_text");

	new KAction(i18n("Hatches"), 
		"qg_menuhatch", 0, actionHandler, SLOT(slotDrawHatch()),
		actionCollection(), "draw_hatch");

	new KAction(i18n("Image"), "qg_menuimage",
		0, actionHandler, SLOT(slotDrawImage()),
		actionCollection(), "draw_image");

	menu = new KActionMenu(i18n("Dimensions"), "qg_dimaligned",
		actionCollection(), "draw_dimensions");
	menu->setDelayed(true);
	toggle = new KToggleAction(i18n("Alligned Dimension"), 
		"qg_dimaligned", 0, actionHandler, SLOT(slotDimAligned()),
		actionCollection(), "dim_aligned");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Horizontal Dimensions"), 
		"qg_dimhor", 0, actionHandler,
		SLOT(slotDimLinearHor()),actionCollection(), "dim_horizontal");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Vertical Dimensions"), 
		"qg_dimver", 0, actionHandler,
		SLOT(slotDimLinearVer()),
		actionCollection(), "dim_vertical");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Radial Dimension"),
		"qg_dimrad", 0, actionHandler,
		SLOT(slotDimRadial()), actionCollection(), "dim_radial");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Diametric Dimension"), 
		"qg_dimdiametric", 0, actionHandler,
		SLOT(slotDimDiametric()),
		actionCollection(), "dim_diametric");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Angular Dimension"),
		"qg_dimangular", 0, actionHandler, SLOT(slotDimAngular()),
		actionCollection(), "dim_angular");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);
	toggle = new KToggleAction(i18n("Leader"),
		"qg_dimleader", 0, actionHandler, SLOT(slotDimLeader()),
		actionCollection(), "dim_leader");
	toggle->setExclusiveGroup("draw");
	menu->insert(toggle);

	menu = new KActionMenu(i18n("Modify"), "qg_menuedit",
		actionCollection(), "modify");
	menu->setDelayed(true);
	action = new KAction(i18n("Move"), "qg_modifymove",
		0, actionHandler, SLOT(slotModifyMove()),
		actionCollection(), "modify_move");
	menu->insert(action);
	action = new KAction(i18n("Rotate"), "qg_modifyrotate",
		0, actionHandler, SLOT(slotModifyRotate()),
		actionCollection(), "modify_rotate");
	menu->insert(action);
	action = new KAction(i18n("Scale"), "qg_modifyscale",
		0, actionHandler, SLOT(slotModifyScale()),
		actionCollection(), "modify_scale");
	menu->insert(action);
	action = new KAction(i18n("Mirror"), "qg_modifymirror",
		0, actionHandler, SLOT(slotModifyMirror()),
		actionCollection(), "");
	menu->insert(action);
	action = new KAction(i18n("Move and Rotate"),
		"qg_modifymoverotate", 0, actionHandler,
		SLOT(slotModifyMoveRotate()),
		actionCollection(), "modify_move_rotate");
	menu->insert(action);
	action = new KAction(i18n("Rotate around two centers"),
		"qg_modifyrotate2", 0, actionHandler,
		SLOT(slotModifyRotate2()),
		actionCollection(), "modify_rotate_2");
	menu->insert(action);
	action = new KAction(i18n("Trim"), "qg_modifytrim",
		0, actionHandler, SLOT(slotModifyTrim()),
		actionCollection(), "modify_trim");
	menu->insert(action);
	action = new KAction(i18n("Trim by amount"),
		"qg_modifytrimamount", 0, actionHandler, 
		SLOT(slotModifyTrimAmount()),
		actionCollection(), "modify_trim_amount");
	menu->insert(action);
	action = new KAction(i18n("Stretch"), "qg_modifystretch",
		0, actionHandler, SLOT(slotModifyStretch()),
		actionCollection(), "modify_stretch");
	menu->insert(action);
	action = new KAction(i18n("Bevel"), "qg_modifybevel",
		0, actionHandler, SLOT(slotModifyBevel()),
		actionCollection(), "modify_bevel");
	menu->insert(action);
	action = new KAction(i18n("Round"), "qg_modifyround",
		0, actionHandler, SLOT(slotModifyRound()),
		actionCollection(), "modify_round");
	menu->insert(action);
	action = new KAction(i18n("Divide"), "qg_modifycut",
		0, actionHandler, SLOT(slotModifyCut()),
		actionCollection(), "modify_cut");
	menu->insert(action);
	action = new KAction(i18n("Edit Entity Geometry"), 
		"qg_modifyentity", 0, actionHandler,
		SLOT(slotModifyEntity()),
		actionCollection(), "modify_entity");
	menu->insert(action);
	action = new KAction(i18n("Edit Entity Attributes"),
		"qg_modifyattributes", 0, actionHandler,
		SLOT(slotModifyAttributes()),
		actionCollection(), "modify_attributes");
	menu->insert(action);
	action = new KAction(i18n("Delete"), "qg_modifydelete",
		0, actionHandler, SLOT(slotModifyDelete()),
		actionCollection(), "modify_delete");
	menu->insert(action);
	action = new KAction(i18n("Explode Text into Letters"), 
		"qg_modifyexplodetext", 0, actionHandler,
		SLOT(slotModifyExplodeText()),
		actionCollection(), "modify_explode_text");
	menu->insert(action);

	toggle = new KToggleAction(i18n("Snap Free"), 
		"qg_snapfree", 0, actionHandler, SLOT(slotSnapFree()),
		actionCollection(), "snap_free");
	toggle->setExclusiveGroup("snap");
	toggle->setChecked(true);
	actionHandler->setActionSnapFree(toggle);
	toggle = new KToggleAction(i18n("Snap Grid"), 
		"qg_snapgrid", 0, actionHandler, SLOT(slotSnapGrid()),
		actionCollection(), "snap_grid");
	toggle->setExclusiveGroup("snap");
	actionHandler->setActionSnapGrid(toggle);
	toggle = new KToggleAction(i18n("Snap Endpoint"), 
		"qg_snapendpoint", 0, actionHandler,
		SLOT(slotSnapEndpoint()),
		actionCollection(), "snap_endpoint");
	toggle->setExclusiveGroup("snap");
	actionHandler->setActionSnapEndpoint(toggle);
	toggle = new KToggleAction(i18n("Snap on Entity"),
		"qg_snaponentity", 0, actionHandler,
		SLOT(slotSnapOnEntity()),
		actionCollection(), "snap_on_entity");
	toggle->setExclusiveGroup("snap");
	actionHandler->setActionSnapOnEntity(toggle);
	toggle = new KToggleAction(i18n("Snap Center"),
		"qg_snapcenter", 0, actionHandler, SLOT(slotSnapCenter()),
		actionCollection(), "snap_center");
	toggle->setExclusiveGroup("snap");
	actionHandler->setActionSnapCenter(toggle);
	toggle = new KToggleAction(i18n("Snap Middle"), 
		"qg_snapmiddle", 0, actionHandler, SLOT(slotSnapMiddle()),
		actionCollection(), "snap_middle");
	toggle->setExclusiveGroup("snap");
	actionHandler->setActionSnapMiddle(toggle);
	toggle = new KToggleAction(i18n("Snap with Distance"), 
		"qg_snapdist", 0, actionHandler, SLOT(slotSnapDist()),
		actionCollection(), "snap_distance");
	toggle->setExclusiveGroup("snap");
	actionHandler->setActionSnapDist(toggle);
	toggle = new KToggleAction(i18n("Snap Intersection"), 
		"qg_snapintersection", 0, actionHandler,
		SLOT(slotSnapIntersection()),
		actionCollection(), "snap_intersection");
	toggle->setExclusiveGroup("snap");
	actionHandler->setActionSnapIntersection(toggle);
	toggle = new KToggleAction(i18n("Snap Intersection Manually"), 
		"qg_snapintersectionm", 0, actionHandler,
		SLOT(slotSnapIntersectionManual()),
		actionCollection(), "snap_intersection_manual");
	toggle->setExclusiveGroup("snap");
	actionHandler->setActionSnapIntersectionManual(toggle);

	toggle = new KToggleAction(i18n("No Restriction"), 
		"qg_restrictnothing", 0, actionHandler,
		SLOT(slotRestrictNothing()),
		actionCollection(), "restrict_nothing");
	toggle->setExclusiveGroup("restrict");
	toggle->setChecked(true);
	actionHandler->setActionRestrictNothing(toggle);
	toggle = new KToggleAction(i18n("Orthogonal Restriction"), 
		"qg_restrictorthogonal", 0, actionHandler,
		SLOT(slotRestrictOrthogonal()),
		actionCollection(), "restrict_orthogonal");
	toggle->setExclusiveGroup("restrict");
	actionHandler->setActionRestrictOrthogonal(toggle);
	toggle = new KToggleAction(i18n("Horizontal Restriction"),
		"qg_restricthorizontal", 0, actionHandler,
		SLOT(slotRestrictHorizontal()),
		actionCollection(), "restrict_horizontal");
	toggle->setExclusiveGroup("restrict");
	actionHandler->setActionRestrictHorizontal(toggle);
	toggle = new KToggleAction(i18n("Vertical Restriction"),
		"qg_restrictvertical", 0, actionHandler,
		SLOT(slotRestrictVertical()),
		actionCollection(), "restrict_vertical");
	toggle->setExclusiveGroup("restrict");
	actionHandler->setActionRestrictVertical(toggle);

	action = new KAction(i18n("Move Relative Zero"), 
		"qg_relzeromove", 0, actionHandler,
		SLOT(slotSetRelativeZero()),
		actionCollection(), "set_relative_zero");
	toggle = new KToggleAction(i18n("Lock Relative Zero"),
		"qg_relzerolock", 0, actionHandler, 
		0, actionCollection(), "lock_relative_zero");
	actionHandler->setActionLockRelativeZero(toggle);

	menu = new KActionMenu(i18n("Measure"), "qg_menumeasure",
		actionCollection(), "measure");
	menu->setDelayed(true);
	new KAction(i18n("Distance Point to Point"),
		"qg_infodist", 0, actionHandler, SLOT(slotInfoDist()),
		actionCollection(), "measure_distance_point");
	new KAction(i18n("Distance Entity to Point"),
		"qg_infodist2", 0, actionHandler, SLOT(slotInfoDist2()),
		actionCollection(), "measure_distance_entity");
	new KAction(i18n("Angle"),
		"qg_infoangle", 0, actionHandler, SLOT(slotInfoAngle()),
		actionCollection(), "measure_angle");
	new KAction(i18n("Measure Total Length"),
		"qg_infototallength", 0, actionHandler,
		SLOT(slotInfoTotalLength()),
		actionCollection(), "measure_total_length");

	new KAction(i18n("Defreeze All"), 0, actionHandler,
		SLOT(slotLayersDefreezeAll()), 
		actionCollection(), "layers_defreeze_all");
	new KAction(i18n("Freeze All"), 0, actionHandler, SLOT(slotLayersFreezeAll()),
		actionCollection(), "layers_freeze_all");
	new KAction(i18n("Add Layer"), 0, actionHandler, SLOT(slotLayersAdd()),
		actionCollection(), "layers_add");
	new KAction(i18n("Remove Layer"), 0, actionHandler, SLOT(slotLayersRemove()),
		actionCollection(), "layers_remove");
	new KAction(i18n("Edit Layer"), 0, actionHandler, SLOT(slotLayersEdit()),
		actionCollection(), "layers_edit");
	new KAction(i18n("Toggle View"), 0, actionHandler,
		SLOT(slotLayersToggleView()),
		actionCollection(), "layers_view");

	new KAction(i18n("Defreeze All"), 0, actionHandler,
		SLOT(slotBlocksDefreezeAll()),
		actionCollection(), "blocks_defreeze_all");
	new KAction(i18n("Freeze All"), 0, actionHandler,
		SLOT(slotBlocksFreezeAll()),
		actionCollection(), "blocks_freeze_all");
	new KAction(i18n("Add Block"), 0, actionHandler, SLOT(slotBlocksAdd()),
		actionCollection(), "blocks_add");
	new KAction(i18n("Remove Block"), 0, actionHandler, SLOT(slotBlocksRemove()),
		actionCollection(), "blocks_remove");
	new KAction(i18n("Edit Attributes"), 0, actionHandler,
		SLOT(slotBlocksAttributes()),
		actionCollection(), "blocks_attributes");
	new KAction(i18n("Insert Block"), 0, actionHandler, SLOT(slotBlocksInsert()),
		actionCollection(), "blocks_insert");
	new KAction(i18n("Edit Block"), 0, actionHandler, SLOT(slotBlocksEdit()),
		actionCollection(), "blocks_edit");
	new KAction(i18n("Create Block"), "qg_menublock", 0, actionHandler, SLOT(slotBlocksCreate()),
		actionCollection(), "blocks_create");
	new KAction(i18n("Explode Block"), 0, actionHandler, SLOT(slotBlocksExplode()),
		actionCollection(), "blocks_explode");

	penToolBar = new QG_PenToolBar(this, actionCollection());
    	connect(penToolBar, SIGNAL(penChanged(RS_Pen)),
        	this, SLOT(slotPenChanged(RS_Pen)));
}


/**
 * Initializes the menu bar.
 */
void QC_ApplicationWindow::initMenuBar() {
    RS_DEBUG->print("QC_ApplicationWindow::initMenuBar()");
}



/**
 * Initializes the tool bars (file tool bar and pen tool bar).
 */
void QC_ApplicationWindow::initToolBar() {
    RS_DEBUG->print("QC_ApplicationWindow::initToolBar()");


    optionWidget = new KToolBar(this, QMainWindow::Top, false, "Tool Options");
    optionWidget->setFixedExtentHeight(26);
}

/**
 * Initializes the status bar at the bottom.
 */
void QC_ApplicationWindow::initStatusBar() {
    RS_DEBUG->print("QC_ApplicationWindow::initStatusBar()");

    statusBar()->setMinimumHeight(32);
    coordinateWidget = new QG_CoordinateWidget(statusBar(), "coordinates");
    statusBar()->addWidget(coordinateWidget);
    mouseWidget = new QG_MouseWidget(statusBar(), "mouse info");
    statusBar()->addWidget(mouseWidget);
    selectionWidget = new QG_SelectionWidget(statusBar(), "selections");
    statusBar()->addWidget(selectionWidget);
}

/**
 * Initializes the view.
 */
void QC_ApplicationWindow::initView() {
    RS_DEBUG->print("QC_ApplicationWindow::initView()");

    RS_DEBUG->print("init view..");
    QDockWindow* dw;
    layerWidget = NULL;
    blockWidget = NULL;
    libraryWidget = NULL;
    commandWidget = NULL;

    RS_DEBUG->print("  layer widget..");
    dw = new QDockWindow(QDockWindow::InDock, this, "Layer");
    layerWidget = new QG_LayerWidget(actionHandler, dw, "Layer");
    layerWidget->setFocusPolicy(QWidget::NoFocus);
    connect(layerWidget, SIGNAL(escape()),
            this, SLOT(slotFocus()));
    connect(this, SIGNAL(windowsChanged(bool)),
            layerWidget, SLOT(setEnabled(bool)));
    //dw->boxLayout()->addWidget(layerWidget);
    dw->setWidget(layerWidget);
    dw->setFixedExtentWidth(120);
    //dw->setFixedExtentHeight(400);
    //dw->setFixedHeight(400);
    dw->setResizeEnabled(true);
    dw->setCaption(tr("Layer List"));
    dw->setCloseMode(QDockWindow::Always);
    addDockWindow(dw, DockRight);
    layerDockWindow = dw;

    RS_DEBUG->print("  block widget..");
    dw = new QDockWindow(QDockWindow::InDock, this, "Block");
    dw->setResizeEnabled(true);
    blockWidget = new QG_BlockWidget(actionHandler, dw, "Block");
    blockWidget->setFocusPolicy(QWidget::NoFocus);
    connect(blockWidget, SIGNAL(escape()),
            this, SLOT(slotFocus()));
    connect(this, SIGNAL(windowsChanged(bool)),
            blockWidget, SLOT(setEnabled(bool)));
    //dw->boxLayout()->addWidget(blockWidget);
    dw->setWidget(blockWidget);
    dw->setFixedExtentWidth(120);
    dw->setCaption(tr("Block List"));
    dw->setCloseMode(QDockWindow::Always);
    //dw->setFixedExtentHeight(400);
    addDockWindow(dw, DockRight);
    blockDockWindow = dw;

    RS_DEBUG->print("  library widget..");
    dw = new QDockWindow(QDockWindow::OutsideDock, this, "Library");
    libraryWidget = new QG_LibraryWidget(dw, "Library");
    libraryWidget->setActionHandler(actionHandler);
    libraryWidget->setFocusPolicy(QWidget::NoFocus);
    connect(libraryWidget, SIGNAL(escape()),
            this, SLOT(slotFocus()));
    connect(this, SIGNAL(windowsChanged(bool)),
            (QObject*)libraryWidget->bInsert, SLOT(setEnabled(bool)));
    dw->setWidget(libraryWidget);
    //dw->setFixedExtentWidth(240);
    //dw->setHeight(400);
    dw->resize(240, 400);
    dw->setResizeEnabled(true);
    dw->setCaption(tr("Library Browser"));
    dw->setCloseMode(QDockWindow::Always);
    addDockWindow(dw, DockTornOff);
    libraryDockWindow = dw;
    libraryDockWindow->hide();

    RS_DEBUG->print("  command widget..");
    dw = new QDockWindow(QDockWindow::InDock, this, "Command");
    dw->setResizeEnabled(true);
    commandWidget = new QG_CommandWidget(dw, "Command");
    commandWidget->setActionHandler(actionHandler);
    //commandWidget->redirectStderr();
    //std::cerr << "Ready.\n";
    //commandWidget->processStderr();
    connect(this, SIGNAL(windowsChanged(bool)),
            commandWidget, SLOT(setEnabled(bool)));
    //connect(commandWidget, SIGNAL(escape()),
    //        this, SLOT(slotFocus()));
    //commandWidget->grabKeyboard();
    //dw->boxLayout()->addWidget(commandWidget);
    dw->setWidget(commandWidget);
    //dw->setFixedExtentWidth(120);
    dw->setFixedExtentHeight(45);
    dw->setCaption(tr("Command line"));
    dw->setCloseMode(QDockWindow::Always);
    commandDockWindow = dw;
    addDockWindow(dw, DockBottom);

    RS_DEBUG->print("  done");
}

/**
 * Goes back to the previous menu or one step in the current action.
 */
void QC_ApplicationWindow::slotBack() {
    RS_GraphicView* graphicView = getGraphicView();
    if (graphicView) graphicView->back();
}



/**
 * Goes one step further in the current action.
 */
void QC_ApplicationWindow::slotEnter() {

    if (commandWidget==NULL || !commandWidget->checkFocus()) {
            RS_GraphicView* graphicView = getGraphicView();
            if (graphicView) graphicView->enter();
    }
}



/**
 * Sets the keyboard focus on the command line.
 */
void QC_ApplicationWindow::slotFocusCommandLine() {
    if (commandWidget->isVisible()) {
        commandWidget->setFocus();
    }
}
	

/**
 * Shows the given error on the command line.
 */
void QC_ApplicationWindow::slotError(const QString& msg) {
	commandWidget->appendHistory(msg);
}



/**
 * Hands focus back to the application window. In the rare event
 * of a escape press from the layer widget (e.g after switching desktops
 * in XP).
 */
void QC_ApplicationWindow::slotFocus() {
    //QG_GraphicView* graphicView = getGraphicView();
    /*if (graphicView!=NULL) {
        graphicView->setFocus();
}
    else {*/
    setFocus();
    //}
}


/**
 * Called when a document window was activated.
 */
void QC_ApplicationWindow::slotWindowActivated(QWidget*w) {
	QC_GraphicView *view = dynamic_cast<QC_GraphicView*>(w);
	if (!view) return;
	RS_Document *doc = view->getDocument();
	if (!doc) return;
	
        bool showByBlock = doc->rtti()==RS2::EntityBlock;

        layerWidget->setLayerList(doc->getLayerList(),
                                  showByBlock);

        coordinateWidget->setGraphic(doc->getGraphic());

        // Only graphics show blocks. (blocks don't)
        if (doc->rtti()==RS2::EntityGraphic) {
            blockWidget->setBlockList(doc->getBlockList());
        } else {
            blockWidget->setBlockList(0);
        }

        // Update all inserts in this graphic (blocks might have changed):
        doc->updateInserts();
        view->redraw();

        // set snapmode from snapping menu
        actionHandler->updateSnapMode();

        // set pen from pen toolbar
        slotPenChanged(penToolBar->getPen());

        // update toggle button status:
        if (doc->getGraphic()) {
            emit(gridChanged(doc->getGraphic()->isGridOn()));
            emit(printPreviewChanged(view->isPrintPreview()));
        }
    	emit windowsChanged(true);
}

/**
 * CAM
 */
/*
#ifdef RS_CAM
void QC_ApplicationWindow::slotCamExportAuto() {
    printf("CAM export..\n");
    
    RS_Document* d = getDocument();
    if (d!=NULL) {
        RS_Graphic* graphic = (RS_Graphic*)d;
 
        RS_CamDialog dlg(graphic, this);
        dlg.exec();
    }
}
#endif
*/



/**
 * Called when something changed in the pen tool bar 
 * (e.g. color, width, style).
 */
void QC_ApplicationWindow::slotPenChanged(RS_Pen pen) {
    QC_MDIWindow* m = getMDIWindow();
    if (m!=NULL) {
        m->slotPenChanged(pen);
    }
}

/**
 * Creates a new MDI window with the given document or a new
 *  document if 'doc' is NULL.
 */
QC_MDIWindow* QC_ApplicationWindow::slotFileNew(RS_Document* doc) {

    RS_DEBUG->print("QC_ApplicationWindow::slotFileNew() begin");

    static int id = 0;
    id++;

    statusBar()->message(tr("Creating new file..."));

    QC_MDIWindow* w = new QC_MDIWindow(doc, tabWidget());
    createGUI(w);
	//w->setWindowState(WindowMaximized);
//    connect(w, SIGNAL(destroyed()),
 //           this, SLOT(slotFileClosing()));

    if (w->getDocument()->rtti()==RS2::EntityBlock) {
	tabWidget()->addTab(w->widget(), i18n("Block '%1'").
		arg(((RS_Block*)(w->getDocument()))->getName()));
    } else {
	tabWidget()->addTab(w->widget(), i18n("unnamed document %1").arg(id));
    }

    // only graphics offer block lists, blocks don't
    RS_DEBUG->print("  adding listeners");
    RS_Graphic* graphic = w->getDocument()->getGraphic();
    if (graphic!=NULL) {
        // Link the graphic's layer list to the pen tool bar
        graphic->addLayerListListener(penToolBar);
        // Link the layer list to the layer widget
        graphic->addLayerListListener(layerWidget);

        // Link the block list to the block widget
        graphic->addBlockListListener(blockWidget);
    }

    // Link the dialog factory to the mouse widget:
    QG_DIALOGFACTORY->setMouseWidget(mouseWidget);
    // Link the dialog factory to the coordinate widget:
    QG_DIALOGFACTORY->setCoordinateWidget(coordinateWidget);
    QG_DIALOGFACTORY->setSelectionWidget(selectionWidget);
    // Link the dialog factory to the option widget:
    //QG_DIALOGFACTORY->setOptionWidget(optionWidget);
    // Link the dialog factory to the command widget:
    QG_DIALOGFACTORY->setCommandWidget(commandWidget);
    // Link the dialog factory to the main app window:
    QG_DIALOGFACTORY->setMainWindow(this);

    slotWindowActivated(tabWidget()->currentPage());
    statusBar()->message(i18n("New Drawing created."), 2000);

    RS_DEBUG->print("QC_ApplicationWindow::slotFileNew() OK");
    setFocus();
	
    return w;
}



/**
 * Menu file -> open.
 */
void QC_ApplicationWindow::slotFileOpen() {
	KURL url = KFileDialog::getOpenURL(QString::null, "*.dxf|DXF Files");
	if (url.isEmpty()) return;
	slotFileOpen(url);
}

/**
 * Menu file -> open.
 */
void QC_ApplicationWindow::slotFileOpen(KURL url) {
	if (url.isEmpty()) return;
	if (!url.isLocalFile()) return;
	QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );

        // Create new document window:
        QC_MDIWindow* w = slotFileNew();
        RS_APP->processEvents(1000);

        // link the layer widget to the new document:
        layerWidget->setLayerList(w->getDocument()->getLayerList(), false);
        // link the block widget to the new document:
        blockWidget->setBlockList(w->getDocument()->getBlockList());
        // link coordinate widget to graphic
        coordinateWidget->setGraphic(w->getGraphic());

        RS_APP->processEvents(1000);

        // open the file in the new view:
        if (!w->openURL(url)) {
            // error
            QApplication::restoreOverrideCursor();
            QMessageBox::information(this, QMessageBox::tr("Warning"),
                                     tr("Cannot open the file\n%1\nPlease "
                                        "check the permissions.")
                                     .arg(url.url()),
                                     QMessageBox::Ok);
            w->setForceClosing(true);
	    delete w;
            return;
        }

        RS_DEBUG->print("QC_ApplicationWindow::slotFileOpen: open file: OK");

        RS_DEBUG->print("QC_ApplicationWindow::slotFileOpen: update recent file menu: 1");

        // update recent files menu:
	recentFiles->addURL(url);
        RS_DEBUG->print("QC_ApplicationWindow::slotFileOpen: set caption");
        // update caption:
	tabWidget()->changeTab(w->widget(), url.fileName());
        RS_DEBUG->print("QC_ApplicationWindow::slotFileOpen: set caption: OK");

        RS_DEBUG->print("QC_ApplicationWindow::slotFileOpen: update coordinate widget");
        // update coordinate widget format:
        RS_DIALOGFACTORY->updateCoordinateWidget(RS_Vector(0.0,0.0),
                RS_Vector(0.0,0.0),
                true);
        RS_DEBUG->print("QC_ApplicationWindow::slotFileOpen: update coordinate widget: OK");

        // show output of filter (if any):
        commandWidget->processStderr();
        QString message=tr("Loaded document: ")+url.url();
        commandWidget->appendHistory(message);
        statusBar()->message(message, 2000);
    	QApplication::restoreOverrideCursor();
}



/**
 * Menu file -> save.
 */
void QC_ApplicationWindow::slotFileSave() {
    RS_DEBUG->print("QC_ApplicationWindow::slotFileSave()");

    statusBar()->message(tr("Saving drawing..."));

    QC_MDIWindow* w = getMDIWindow();
    QString name;
    if (w!=NULL) {
        if (w->getDocument()->getFilename().isEmpty()) {
            slotFileSaveAs();
        } else {
            bool cancelled;
            if (w->slotFileSave(cancelled)) {
                if (!cancelled) {
                    name = w->getDocument()->getFilename();
                    statusBar()->message(tr("Saved drawing: %1").arg(name), 2000);
                }
            } else {
                // error
                QMessageBox::information(this, QMessageBox::tr("Warning"),
                                         tr("Cannot save the file\n%1\nPlease "
                                            "check the permissions.")
                                         .arg(w->getDocument()->getFilename()),
                                         QMessageBox::Ok);
            }
        }
    }
}



/**
 * Menu file -> save as.
 */
void QC_ApplicationWindow::slotFileSaveAs() {
    RS_DEBUG->print("QC_ApplicationWindow::slotFileSaveAs()");

    statusBar()->message(tr("Saving drawing under new filename..."));

    QC_MDIWindow* w = getMDIWindow();
    QString name;
    if (w!=NULL) {
        bool cancelled;
        if (w->slotFileSaveAs(cancelled)) {
            if (!cancelled) {
            	recentFiles->addURL(w->url());
		tabWidget()->changeTab(w->widget(), w->url().fileName());
	    }
        } else {
            // error
            QMessageBox::information(this, QMessageBox::tr("Warning"),
                                     tr("Cannot save the file\n%1\nPlease "
                                        "check the permissions.")
                                     .arg(w->getDocument()->getFilename()),
                                     QMessageBox::Ok);
        }
    }

    QString message = tr("Saved drawing: %1").arg(name);
    statusBar()->message(message, 2000);
    commandWidget->appendHistory(message);
}



/**
 * Menu file -> export.
 */
void QC_ApplicationWindow::slotFileExport() {
    RS_DEBUG->print("QC_ApplicationWindow::slotFileExport()");

    statusBar()->message(tr("Exporting drawing..."));

    QC_MDIWindow* w = getMDIWindow();
    QString fn;
    if (w!=NULL) {

        // read default settings:
        RS_SETTINGS->beginGroup("/Paths");
        RS_String defDir = RS_SETTINGS->readEntry("/ExportImage",
                           RS_SYSTEM->getHomeDir());
        RS_String defFilter = RS_SETTINGS->readEntry("/ExportImageFilter",
                              "Portable Network Graphic (*.png)");
        RS_SETTINGS->endGroup();

        bool cancel = false;

        QFileDialog fileDlg(NULL, "", true);

        QStrList f = QImageIO::outputFormats();
        QStringList formats = QStringList::fromStrList(f);
        QStringList filters;
        //QString all = "";

        for (QStringList::Iterator it = formats.begin();
                it!=formats.end(); ++it) {

            QString st;
            if ((*it)=="JPEG") {
                st = QString("%1 (*.%2 *.jpg)")
                     .arg(QG_DialogFactory::extToFormat(*it))
                     .arg(QString(*it).lower());
            } else {
                st = QString("%1 (*.%2)")
                     .arg(QG_DialogFactory::extToFormat(*it))
                     .arg(QString(*it).lower());
            }
            filters.append(st);

            //if (!all.isEmpty()) {
            //    all += " ";
            //}
            //all += QString("*.%1").arg(QString(*it).lower());
        }

        fileDlg.setFilters(filters);
        fileDlg.setMode(QFileDialog::AnyFile);
        fileDlg.setCaption(QObject::tr("Export Image"));
        fileDlg.setDir(defDir);
        fileDlg.setSelectedFilter(defFilter);

        if (fileDlg.exec()==QDialog::Accepted) {
            fn = fileDlg.selectedFile();
            cancel = false;
        } else {
            cancel = true;
        }

        // store new default settings:
        if (!cancel) {
            RS_SETTINGS->beginGroup("/Paths");
            RS_SETTINGS->writeEntry("/ExportImage", QFileInfo(fn).dirPath(true));
            RS_SETTINGS->writeEntry("/ExportImageFilter",
                                    fileDlg.selectedFilter());
            RS_SETTINGS->endGroup();

            // find out extension:
            QString filter = fileDlg.selectedFilter();
            QString format = "";
            int i = filter.find("(*.");
            if (i!=-1) {
                int i2 = filter.find(QRegExp("[) ]"), i);
                format = filter.mid(i+3, i2-(i+3));
                format = format.upper();
            }

            // append extension to file:
            if (!QFileInfo(fn).fileName().contains(".")) {
                fn.append("." + format.lower());
            }

            // show options dialog:
            QG_ImageOptionsDialog dlg(this);
            dlg.setGraphicSize(w->getGraphic()->getSize());
            if (dlg.exec()) {
                bool ret = slotFileExport(fn, format, dlg.getSize(),
                                          dlg.isBackgroundBlack());
                if (ret) {
                    QString message = tr("Exported: %1").arg(fn);
                    statusBar()->message(message, 2000);
                    commandWidget->appendHistory(message);
                }
            }
        }
    }

}



/**
 * Exports the drawing as a bitmap.
 *
 * @param name File name.
 * @param format File format (e.g. "png")
 * @param size Size of the bitmap in pixel
 * @param black true: Black background, false: white
 * @param bw true: black/white export, false: color
 */
bool QC_ApplicationWindow::slotFileExport(const QString& name,
        const QString& format, QSize size, bool black, bool bw) {

    QC_MDIWindow* w = getMDIWindow();
    if (w==NULL) {
        RS_DEBUG->print(RS_Debug::D_WARNING,
        	"QC_ApplicationWindow::slotFileExport: "
        	"no window opened");
        return false;
    }

    RS_Graphic* graphic = w->getDocument()->getGraphic();
    if (graphic==NULL) {
        RS_DEBUG->print(RS_Debug::D_WARNING,
        	"QC_ApplicationWindow::slotFileExport: "
        	"no graphic");
        return false;
    }

    statusBar()->message(tr("Exporting..."));
    QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );

    bool ret = false;
    QPixmap* buffer = new QPixmap(size);
    RS_PainterQt* painter = new RS_PainterQt(buffer);

    // black background:
    if (black) {
        painter->setBackgroundColor(RS_Color(0,0,0));
    }
    // white background:
    else {
        painter->setBackgroundColor(RS_Color(255,255,255));
    }

    // black/white:
    if (bw) {
        painter->setDrawingMode(RS2::ModeBW);
    }

    painter->eraseRect(0,0, size.width(), size.height());

    RS_StaticGraphicView gv(size.width(), size.height(), painter);
    if (black) {
        gv.setBackground(RS_Color(0,0,0));
    } else {
        gv.setBackground(RS_Color(255,255,255));
    }
    gv.setContainer(graphic);
    gv.zoomAuto(false);
    gv.drawEntity(graphic, true);

    QImageIO iio;
    QImage img;
    img = *buffer;
    iio.setImage(img);
    iio.setFileName(name);
    iio.setFormat(format);
    if (iio.write()) {
        ret = true;
    }
    QApplication::restoreOverrideCursor();

    // GraphicView deletes painter
    painter->end();
    delete buffer;

    if (ret) {
        statusBar()->message(tr("Export complete"), 2000);
    } else {
        statusBar()->message(tr("Export failed!"), 2000);
    }

    return ret;
}


/**
 * Menu file -> close.
 */
void QC_ApplicationWindow::slotFileClose() {
    QC_MDIWindow* m = getMDIWindow();
    m->close();
}


/**
 * Called when a MDI window is actually about to close. Used to 
 * detach widgets from the document.
 */
void QC_ApplicationWindow::slotFileClosing() {
    layerWidget->setLayerList(NULL, false);
    blockWidget->setBlockList(NULL);
    coordinateWidget->setGraphic(NULL);
#ifdef RS_CAM
    simulationControls->setGraphicView(NULL);
#endif
}



/**
 * Menu file -> print.
 */
void QC_ApplicationWindow::slotFilePrint() {
    RS_DEBUG->print("QC_ApplicationWindow::slotFilePrint()");

    QC_MDIWindow* w = getMDIWindow();
    if (w==NULL) {
        RS_DEBUG->print(RS_Debug::D_WARNING,
        	"QC_ApplicationWindow::slotFilePrint: "
        	"no window opened");
        return;
    }

    RS_Graphic* graphic = w->getDocument()->getGraphic();
    if (graphic==NULL) {
        RS_DEBUG->print(RS_Debug::D_WARNING,
        	"QC_ApplicationWindow::slotFilePrint: "
        	"no graphic");
        return;
    }

    statusBar()->message(tr("Printing..."));
    KPrinter* printer;
    printer = new KPrinter(true, QPrinter::HighResolution);
    bool landscape = false;
    printer->setPageSize((KPrinter::PageSize)RS2::rsToQtPaperFormat(graphic->getPaperFormat(&landscape)));
    if (landscape) {
        printer->setOrientation(KPrinter::Landscape);
    } else {
        printer->setOrientation(KPrinter::Portrait);
    }

    printer->setOutputFileName(RS_SETTINGS->readEntry("/FileName", ""));
    printer->setColorMode((KPrinter::ColorMode)RS_SETTINGS->readNumEntry("/ColorMode", (int)QPrinter::Color));
    printer->setOutputToFile((bool)RS_SETTINGS->readNumEntry("/PrintToFile",
                             0));
    // printer setup:
    if (printer->setup(this)) {
        QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );
        printer->setFullPage(true);
        QPaintDeviceMetrics metr(printer);

        RS_PainterQt* painter = new RS_PainterQt(printer);
        painter->setDrawingMode(w->getGraphicView()->getDrawingMode());

        RS_StaticGraphicView gv(metr.width(), metr.height(), painter);
        gv.setPrinting(true);
        gv.setBorders(0,0,0,0);

        double fx = (double)metr.width() / metr.widthMM()
                    * RS_Units::getFactorToMM(graphic->getUnit());
        double fy = (double)metr.height() / metr.heightMM()
                    * RS_Units::getFactorToMM(graphic->getUnit());

        double f = (fx+fy)/2;

        double scale = graphic->getPaperScale();

        gv.setOffset((int)(graphic->getPaperInsertionBase().x * f),
                     (int)(graphic->getPaperInsertionBase().y * f));
        gv.setFactor(f*scale);

        gv.setContainer(graphic);

        gv.drawEntity(graphic, true);

        // GraphicView deletes painter
        painter->end();

        RS_SETTINGS->beginGroup("/Print");
        RS_SETTINGS->writeEntry("/PrintToFile", (int)printer->outputToFile());
        RS_SETTINGS->writeEntry("/ColorMode", (int)printer->colorMode());
        RS_SETTINGS->writeEntry("/FileName", printer->outputFileName());
        RS_SETTINGS->endGroup();
        QApplication::restoreOverrideCursor();
    }

    delete printer;

    statusBar()->message(tr("Printing complete"), 2000);
}



/**
 * Menu file -> print preview.
 */
void QC_ApplicationWindow::slotFilePrintPreview(bool on) {
    RS_DEBUG->print("QC_ApplicationWindow::slotFilePrintPreview()");

    RS_DEBUG->print("  creating MDI window");
    QC_MDIWindow* parent = getMDIWindow();
    if (parent==NULL) {
        RS_DEBUG->print(RS_Debug::D_WARNING,
        	"QC_ApplicationWindow::slotFilePrintPreview: "
        	"no window opened");
        return;
    }

    // close print preview:
    if (!on) {
        RS_DEBUG->print("QC_ApplicationWindow::slotFilePrintPreview(): off");
        if (parent->getGraphicView()->isPrintPreview()) {
            RS_DEBUG->print("QC_ApplicationWindow::slotFilePrintPreview(): close");
            slotFileClose();
        }
    }

    // open print preview:
    else {
        // look for an existing print preview:
        QC_MDIWindow* ppv = parent->getPrintPreview();
        if (ppv!=NULL) {
            //slotWindowActivated(ppv->widget());
        } else {
            if (!parent->getGraphicView()->isPrintPreview()) {
                QC_MDIWindow* w = new QC_MDIWindow(parent->getGraphic(),
			tabWidget());
		createGUI(w);
                parent->addChildWindow(w);
                //connect(w, SIGNAL(signalClosing()),
                //        this, SLOT(slotFileClosing()));

		tabWidget()->addTab(w->widget(),i18n("Print preview for %1")
			.arg(parent->url().fileName()));
                w->getGraphicView()->setPrintPreview(true);
                w->getGraphicView()->setBackground(RS_Color(255,255,255));
                w->getGraphicView()->setDefaultAction(
                    new RS_ActionPrintPreview(*w->getDocument(), *w->getGraphicView()));

                // only graphics offer block lists, blocks don't
                RS_DEBUG->print("  adding listeners");
                RS_Graphic* graphic = w->getDocument()->getGraphic();
                if (graphic!=NULL) {
                    // Link the layer list to the pen tool bar
                    graphic->addLayerListListener(penToolBar);
                    // Link the layer list to the layer widget
                    graphic->addLayerListListener(layerWidget);

                    // Link the block list to the block widget
                    graphic->addBlockListListener(blockWidget);

                    // Center by default:
                    graphic->centerToPage();
                }

                // Link the graphic view to the mouse widget:
                QG_DIALOGFACTORY->setMouseWidget(mouseWidget);
                // Link the graphic view to the coordinate widget:
                QG_DIALOGFACTORY->setCoordinateWidget(coordinateWidget);
                QG_DIALOGFACTORY->setSelectionWidget(selectionWidget);
                // Link the graphic view to the option widget:
                //QG_DIALOGFACTORY->setOptionWidget(optionWidget);
                // Link the graphic view to the command widget:
                QG_DIALOGFACTORY->setCommandWidget(commandWidget);

                w->getGraphicView()->zoomPage();
                setFocus();

                //slotWindowActivated(tabWidget()->currentPage());
            }
        }
    }
}



/**
 * Menu file -> quit.
 */
void QC_ApplicationWindow::slotFileQuit() {
    statusBar()->message(tr("Exiting application..."));
    if (queryClose()) qApp->exit(0);
}


/**
 * Shows / hides the grid.
 *
 * @param toggle true: show, false: hide.
 */
void QC_ApplicationWindow::slotViewGrid(bool toggle) {
    RS_DEBUG->print("QC_ApplicationWindow::slotViewGrid()");

    QC_MDIWindow* m = getMDIWindow();
    if (m!=NULL) {
        RS_Graphic* g = m->getGraphic();
        if (g!=NULL) {
            g->setGridOn(toggle);
        }
    	updateGrids();
	m->getGraphicView()->repaint();
    }

    RS_DEBUG->print("QC_ApplicationWindow::slotViewGrid() OK");
}



/**
 * Enables / disables the draft mode.
 *
 * @param toggle true: enable, false: disable.
 */
void QC_ApplicationWindow::slotViewDraft(bool toggle) {
    RS_DEBUG->print("QC_ApplicationWindow::slotViewDraft()");

    /*QC_MDIWindow* m = getMDIWindow();
    if (m!=NULL) {
        RS_Graphic* g = m->getGraphic();
        if (g!=NULL) {
            g->setDraftOn(toggle);
        }
}*/

    RS_SETTINGS->beginGroup("/Appearance");
    RS_SETTINGS->writeEntry("/DraftMode", (int)toggle);
    RS_SETTINGS->endGroup();
}

/**
 * Updates all grids of all graphic views.
 */
void QC_ApplicationWindow::updateGrids() {
        for (int i = 0; i < tabWidget()->count(); i++) {
                QG_GraphicView* gv;
                if (gv = dynamic_cast<QG_GraphicView *>(tabWidget()->page(i)))
                    gv->updateGrid();
        }
}



/**
 * Shows / hides the status bar.
 *
 * @param toggle true: show, false: hide.
 */
void QC_ApplicationWindow::slotViewStatusBar(bool toggle) {
    RS_DEBUG->print("QC_ApplicationWindow::slotViewStatusBar()");

    if (toggle==false) {
        statusBar()->hide();
    } else {
        statusBar()->show();
    }
}



/**
 * Shows / hides the layer list.
 *
 * @param toggle true: show, false: hide.
 */
/*void QC_ApplicationWindow::slotViewLayerList(bool toggle) {
    RS_DEBUG->print("QC_ApplicationWindow::slotViewLayerList()");
 
    if (toggle==false) {
        layerDockWindow->hide();
    } else {
        layerDockWindow->show();
    }
}
*/


/**
 * Shows / hides the block list.
 *
 * @param toggle true: show, false: hide.
 */
/*
void QC_ApplicationWindow::slotViewBlockList(bool toggle) {
    RS_DEBUG->print("QC_ApplicationWindow::slotViewBlockList()");
 
    if (toggle==false) {
        blockDockWindow->hide();
    } else {
        blockDockWindow->show();
    }
}
*/



/**
 * Shows / hides the command line.
 *
 * @param toggle true: show, false: hide.
 */
/*
void QC_ApplicationWindow::slotViewCommandLine(bool toggle) {
    RS_DEBUG->print("QC_ApplicationWindow::slotViewCommandLine()");
 
    if (toggle==false) {
        commandDockWindow->hide();
        //QG_GraphicView* graphicView = getGraphicView();
        //if (graphicView!=NULL) {
        //graphicView->setFocus();
        //}
        setFocus();
    } else {
        commandDockWindow->show();
    }
}
*/



/**
 * Shows / hides the option toolbar.
 *
 * @param toggle true: show, false: hide.
 */
/*
void QC_ApplicationWindow::slotViewOptionToolbar(bool toggle) {
    RS_DEBUG->print("QC_ApplicationWindow::slotViewOptionToolbar()");
 
    if (toggle==false) {
        optionWidget->hide();
    } else {
        optionWidget->show();
    }
}
*/


/**
 * Creates a new MDI window for editing the selected block.
 */
/*
void QC_ApplicationWindow::slotBlocksEdit() {
    RS_DEBUG->print("QC_ApplicationWindow::slotBlocksEdit()");
 
    QC_MDIWindow* parent = getMDIWindow();
    if (parent!=NULL) {
        RS_BlockList* blist = blockWidget->getBlockList();
        if (blist!=NULL) {
            RS_Block* blk = blist->getActiveBlock();
            if (blk!=NULL) {
                QC_MDIWindow* w = slotFileNew(blk);
                // the parent needs a pointer to the block window and
                //   vice versa
                parent->addChildWindow(w);
                w->getGraphicView()->zoomAuto();
            }
        }
    }
}
*/



/**
 * Shows the dialog for general application preferences.
 */
void QC_ApplicationWindow::slotOptionsGeneral() {
    RS_DIALOGFACTORY->requestOptionsGeneralDialog();

    // update background color of all open drawings:
    RS_SETTINGS->beginGroup("/Appearance");
    QColor color(RS_SETTINGS->readEntry("/BackgroundColor", "#000000"));
    QColor gridColor(RS_SETTINGS->readEntry("/GridColor", "Gray"));
    QColor metaGridColor(RS_SETTINGS->readEntry("/MetaGridColor", "Darkgray"));
    QColor selectedColor(RS_SETTINGS->readEntry("/SelectedColor", "#A54747"));
    QColor highlightedColor(RS_SETTINGS->readEntry("/HighlightedColor",
                            "#739373"));
    RS_SETTINGS->endGroup();

    for (int i = 0; i < tabWidget()->count(); i++) {
            QG_GraphicView* gv;
            if (gv = dynamic_cast<QG_GraphicView*>(tabWidget()->page(i))) {
                gv->setBackground(color);
                gv->setGridColor(gridColor);
                gv->setMetaGridColor(metaGridColor);
                gv->setSelectedColor(selectedColor);
                gv->setHighlightedColor(highlightedColor);
                gv->updateGrid();
                gv->redraw();
            }
    }

}



/**
 * Menu script -> show ide
 */
void QC_ApplicationWindow::slotScriptOpenIDE() {
#ifdef RS_SCRIPTING
    scripter->openIDE();
#endif
}



/**
 * Menu script -> run
 */
void QC_ApplicationWindow::slotScriptRun() {
#ifdef RS_SCRIPTING
    scripter->runScript();
#endif
}



/**
 * Menu help -> about.
 */
void QC_ApplicationWindow::slotHelpAbout() {
    RS_DEBUG->print("QC_ApplicationWindow::slotHelpAbout()");

	QString edition;

#ifdef RS_DEMO
    //demo; // = " [Demo]";
#else
#ifdef RS_PROF
    edition = " [Professional]";
#else
    edition = " [Community Edition]";
#endif
#endif

    QStringList modules;

#ifdef RS_CAM
    modules += "CAM";
#endif

#ifdef RS_SCRIPTING
    modules += "Scripting";
#endif

    QString modulesString;

    if (modules.empty()==false) {
        modulesString = modules.join(", ");
    } else {
        modulesString = tr("None");
    }

    QMessageBox box(this);
    box.setCaption(tr("About..."));
    box.setText(       QString("<qt>") +  // no center for main stream QCad
#ifdef QC_ABOUT_HEADER
                       QString("<center>") + 
                       QString(QC_ABOUT_HEADER) +
#else
                       "<h2>"+ QC_APPNAME+ "</h2>" +
#endif
                       tr("Version: %1 %2").arg(QC_VERSION).arg(edition) + "<br>" +
                       tr("Date: %1").arg(__DATE__) + "<br>" +
                       QString("(c) 2001-2005 by RibbonSoft,<br> Andrew Mustun") + 
                       "<br>" +
                       tr("Modules: %1").arg(modulesString) + "<br>" +
                       QString("http://www.ribbonsoft.com")
#ifdef QC_ABOUT_ADD_COMPANY
                       + QString("<br>") + QC_ABOUT_ADD_COMPANY
                       + QString("</center>")
#endif
                       );
#ifndef QC_ABOUT_HEADER
    box.setIcon( QPixmap::fromMimeSource(QC_APP_ICON) );
    box.setFixedWidth(340);
    box.setFixedHeight(250);
#endif
    box.exec();
}



/**
 * Menu help -> help.
 */
void QC_ApplicationWindow::slotHelpManual() {
    RS_DEBUG->print("QC_ApplicationWindow::slotHelpManual()");

    if (assistant==NULL) {
        RS_DEBUG->print("QC_ApplicationWindow::slotHelpManual(): appdir: %s",
                        RS_SYSTEM->getAppDir().latin1());
        RS_DEBUG->print("QC_ApplicationWindow::slotHelpManual(): appdir: %s",
                        RS_SYSTEM->getAppDir().latin1());
        assistant = new QAssistantClient(RS_SYSTEM->getAppDir()+"/bin", this);
		connect(assistant, SIGNAL(error(const QString&)), 
			this, SLOT(slotError(const QString&)));
        QStringList args;
        args << "-profile";
        args << QDir::convertSeparators(RS_SYSTEM->getDocPath() + "/qcaddoc.adp");
//        args << QString("doc") + QDir::separator() + QString("qcaddoc.adp");

#if QT_VERSION>=0x030200 
        assistant->setArguments(args);
#endif
    }
    assistant->openAssistant();
    //assistant->showPage("index.html");
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestDumpEntities(RS_EntityContainer* d) {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestDumpEntities()");
    static int level = 0;
    std::ofstream dumpFile;

    if (d==NULL) {
        d = getDocument();
        dumpFile.open("debug_entities.html");
        level = 0;
    } else {
        dumpFile.open("debug_entities.html", std::ios::app);
        level++;
    }

    if (d!=NULL) {
        if (level==0) {
            dumpFile << "<html>\n";
            dumpFile << "<body>\n";
        }

        for (RS_Entity* e=d->firstEntity();
                e!=NULL;
                e=d->nextEntity()) {

            dumpFile << "<table border=\"1\">\n";
            dumpFile << "<tr><td>Entity: " << e->getId()
            << "</td></tr>\n";

            dumpFile
            << "<tr><td><table><tr>"
            << "<td>VIS:" << e->isVisible() << "</td>"
            << "<td>UND:" << e->isUndone() << "</td>"
            << "<td>SEL:" << e->isSelected() << "</td>"
            << "<td>TMP:" << e->getFlag(RS2::FlagTemp) << "</td>";
            RS_String lay = "NULL";
            if (e->getLayer()!=NULL) {
                lay = e->getLayer()->getName();
            }
            dumpFile
            << "<td>Layer: " << lay << "</td>"
            << "<td>Width: " << (int)e->getPen(false).getWidth() << "</td>"
            << "<td>Parent: " << e->getParent()->getId() << "</td>"
            << "</tr></table>";

            dumpFile
            << "<tr><td>\n";

            switch (e->rtti()) {
            case RS2::EntityPoint: {
                    RS_Point* p = (RS_Point*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Point:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>"
                    << p->getPos()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            case RS2::EntityLine: {
                    RS_Line* l = (RS_Line*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Line:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>"
                    << l->getStartpoint()
                    << "</td>"
                    << "<td>"
                    << l->getEndpoint()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            case RS2::EntityArc: {
                    RS_Arc* a = (RS_Arc*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Arc:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>Center: "
                    << a->getCenter()
                    << "</td>"
                    << "<td>Radius: "
                    << a->getRadius()
                    << "</td>"
                    << "<td>Angle 1: "
                    << a->getAngle1()
                    << "</td>"
                    << "<td>Angle 2: "
                    << a->getAngle2()
                    << "</td>"
                    << "<td>Startpoint: "
                    << a->getStartpoint()
                    << "</td>"
                    << "<td>Endpoint: "
                    << a->getEndpoint()
                    << "</td>"
                    << "<td>reversed: "
                    << (int)a->isReversed()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            case RS2::EntityCircle: {
                    RS_Circle* c = (RS_Circle*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Circle:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>Center: "
                    << c->getCenter()
                    << "</td>"
                    << "<td>Radius: "
                    << c->getRadius()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            case RS2::EntityDimAligned: {
                    RS_DimAligned* d = (RS_DimAligned*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Dimension / Aligned:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>"
                    << d->getDefinitionPoint()
                    << "</td>"
                    << "<td>"
                    << d->getExtensionPoint1()
                    << "</td>"
                    << "<td>"
                    << d->getExtensionPoint2()
                    << "</td>"
                    << "<td>Text: "
                    << d->getText().latin1()
                    << "</td>"
                    << "<td>Label: "
                    << d->getLabel().latin1()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            case RS2::EntityDimLinear: {
                    RS_DimLinear* d = (RS_DimLinear*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Dimension / Linear:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>"
                    << d->getDefinitionPoint()
                    << "</td>"
                    << "<td>"
                    << d->getExtensionPoint1()
                    << "</td>"
                    << "<td>"
                    << d->getExtensionPoint2()
                    << "</td>"
                    << "<td>Text: "
                    << d->getText()
                    << "</td>"
                    << "<td>Label: "
                    << d->getLabel()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            case RS2::EntityInsert: {
                    RS_Insert* i = (RS_Insert*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Insert:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>Insertion point:"
                    << i->getInsertionPoint()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            case RS2::EntityText: {
                    RS_Text* t = (RS_Text*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Text:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>Text:"
                    << t->getText().latin1()
                    << "</td>"
                    << "<td>Height:"
                    << t->getHeight()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            case RS2::EntityHatch: {
                    RS_Hatch* h = (RS_Hatch*)e;
                    dumpFile
                    << "<table><tr><td>"
                    << "<b>Hatch:</b>"
                    << "</td></tr>";
                    dumpFile
                    << "<tr>"
                    << "<td>Pattern:"
                    << h->getPattern().latin1()
                    << "</td>"
                    << "<td>Scale:"
                    << h->getScale()
                    << "</td>"
                    << "<td>Solid:"
                    << (int)h->isSolid()
                    << "</td>"
                    << "</tr></table>";
                }
                break;

            default:
                dumpFile
                << "<tr><td>"
                << "<b>Unknown Entity: " << e->rtti() << "</b>"
                << "</td></tr>";
                break;
            }

            if (e->isContainer() || e->rtti()==RS2::EntityHatch) {
                RS_EntityContainer* ec = (RS_EntityContainer*)e;
                dumpFile << "<table><tr><td valign=\"top\">&nbsp;&nbsp;&nbsp;&nbsp;Contents:</td><td>\n";
                dumpFile.close();
                slotTestDumpEntities(ec);
                dumpFile.open("debug_entities.html", std::ios::app);
                dumpFile << "</td></tr></table>\n";
            }

            dumpFile
            << "</td></tr>"
            << "</table>\n"
            << "<br><br>";
        }

        if (level==0) {
            dumpFile << "</body>\n";
            dumpFile << "</html>\n";
        } else {
            level--;
        }
    }
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestDumpUndo() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestDumpUndo()");

    RS_Document* d = getDocument();
	if (d!=NULL) {
		std::cout << *(RS_Undo*)d;
		std::cout << std::endl;
	}
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestUpdateInserts() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestUpdateInserts()");

    RS_Document* d = getDocument();
    if (d!=NULL) {
        d->updateInserts();
    }
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestDrawFreehand() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestDrawFreehand()");


    //RS_Graphic* g = document->getMarking();
    /*

       RS_ActionDrawLineFree* action =
          new RS_ActionDrawLineFree(*document->getGraphic(),
                                    *graphicView);

       for (int i=0; i<100; ++i) {

           int posx = (random()%600);
           int posy = (random()%400);

           //RS_MouseEvent rsm1(posx, posy, LEFT);
        RS_MouseEvent rsm1(QEvent::MouseButtonPress, 
                           QPoint(posx,posy), 
                           RS2::LeftButton,
                           RS2::LeftButton);
           action->mousePressEvent(&rsm1);

           int speedx = 0;
           int speedy = 0;

           for (int k=0; k<100; ++k) {
               int accx = (random()%40)-20;
               int accy = (random()%40)-20;

               speedx+=accx;
               speedy+=accy;

               posx+=speedx;
               posy+=speedy;

               //RS_MouseEvent rsm2(posx, posy, LEFT);
            
            RS_MouseEvent rsm2(QEvent::MouseMove, 
                           QPoint(posx,posy), 
                           RS2::LeftButton,
                           RS2::LeftButton);
               action->mouseMoveEvent(&rsm2);
           }

           action->mouseReleaseEvent(NULL);

           slotFileSave();
       }

       delete action;
    */
}

QC_MDIWindow *QC_ApplicationWindow::getMDIWindow()
{
	QC_GraphicView *view;
	if (view = dynamic_cast<QC_GraphicView*>(tabWidget()->currentPage()))
		return view->getParent();
	else return 0;
}

/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestInsertBlock() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestInsertBlock()");

    RS_Document* d = getDocument();
    if (d!=NULL && d->rtti()==RS2::EntityGraphic) {
        RS_Graphic* graphic = (RS_Graphic*)d;
        if (graphic==NULL) {
            return;
        }

        graphic->addLayer(new RS_Layer("default"));
        RS_Block* block = new RS_Block(graphic, RS_BlockData("debugblock",
                                       RS_Vector(0.0,0.0), true));

        RS_Line* line;
        RS_Arc* arc;
        RS_Circle* circle;

        // Add one red line:
        line = new RS_Line(block,
                           RS_LineData(RS_Vector(0.0,0.0),
                                       RS_Vector(50.0,0.0)));
        line->setLayerToActive();
        line->setPen(RS_Pen(RS_Color(255, 0, 0),
                            RS2::Width01,
                            RS2::SolidLine));
        block->addEntity(line);

        // Add one line with attributes from block:
        line = new RS_Line(block,
                           RS_LineData(RS_Vector(50.0,0.0),
                                       RS_Vector(50.0,50.0)));
        line->setPen(RS_Pen(RS_Color(RS2::FlagByBlock),
                            RS2::WidthByBlock,
                            RS2::LineByBlock));
        block->addEntity(line);

        // Add one arc with attributes from block:
        RS_ArcData d(RS_Vector(50.0,0.0),
                     50.0, M_PI/2.0, M_PI,
                     false);
        arc = new RS_Arc(block, d);
        arc->setPen(RS_Pen(RS_Color(RS2::FlagByBlock),
                           RS2::WidthByBlock,
                           RS2::LineByBlock));
        block->addEntity(arc);

        // Add one blue circle:
        RS_CircleData circleData(RS_Vector(20.0,15.0),
                                 12.5);
        circle = new RS_Circle(block, circleData);
        circle->setLayerToActive();
        circle->setPen(RS_Pen(RS_Color(0, 0, 255),
                              RS2::Width01,
                              RS2::SolidLine));
        block->addEntity(circle);


        graphic->addBlock(block);



        RS_Insert* ins;
        RS_InsertData insData("debugblock",
                              RS_Vector(0.0,0.0),
                              RS_Vector(1.0,1.0), 0.0,
                              1, 1, RS_Vector(0.0, 0.0),
                              NULL, RS2::NoUpdate);

        // insert one magenta instance of the block (original):
        ins = new RS_Insert(graphic, insData);
        ins->setLayerToActive();
        ins->setPen(RS_Pen(RS_Color(255, 0, 255),
                           RS2::Width02,
                           RS2::SolidLine));
        ins->update();
        graphic->addEntity(ins);

        // insert one green instance of the block (rotate):
        insData = RS_InsertData("debugblock",
                                RS_Vector(-50.0,20.0),
                                RS_Vector(1.0,1.0), 30.0/ARAD,
                                1, 1, RS_Vector(0.0, 0.0),
                                NULL, RS2::NoUpdate);
        ins = new RS_Insert(graphic, insData);
        ins->setLayerToActive();
        ins->setPen(RS_Pen(RS_Color(0, 255, 0),
                           RS2::Width02,
                           RS2::SolidLine));
        ins->update();
        graphic->addEntity(ins);

        // insert one cyan instance of the block (move):
        insData = RS_InsertData("debugblock",
                                RS_Vector(10.0,20.0),
                                RS_Vector(1.0,1.0), 0.0,
                                1, 1, RS_Vector(0.0, 0.0),
                                NULL, RS2::NoUpdate);
        ins = new RS_Insert(graphic, insData);
        ins->setLayerToActive();
        ins->setPen(RS_Pen(RS_Color(0, 255, 255),
                           RS2::Width02,
                           RS2::SolidLine));
        ins->update();
        graphic->addEntity(ins);

        // insert one blue instance of the block:
        for (double a=0.0; a<360.0; a+=45.0) {
            insData = RS_InsertData("debugblock",
                                    RS_Vector(60.0,0.0),
                                    RS_Vector(2.0/5,2.0/5), a/ARAD,
                                    1, 1, RS_Vector(0.0, 0.0),
                                    NULL, RS2::NoUpdate);
            ins = new RS_Insert(graphic, insData);
            ins->setLayerToActive();
            ins->setPen(RS_Pen(RS_Color(0, 0, 255),
                               RS2::Width05,
                               RS2::SolidLine));
            ins->update();
            graphic->addEntity(ins);
        }

        // insert an array of yellow instances of the block:
        insData = RS_InsertData("debugblock",
                                RS_Vector(-100.0,-100.0),
                                RS_Vector(0.2,0.2), M_PI/6.0,
                                6, 4, RS_Vector(100.0, 100.0),
                                NULL, RS2::NoUpdate);
        ins = new RS_Insert(graphic, insData);
        ins->setLayerToActive();
        ins->setPen(RS_Pen(RS_Color(255, 255, 0),
                           RS2::Width01,
                           RS2::SolidLine));
        ins->update();
        graphic->addEntity(ins);


        RS_GraphicView* v = getGraphicView();
        if (v!=NULL) {
            v->redraw();
        }
    }
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestInsertEllipse() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestInsertEllipse()");


    RS_Document* d = getDocument();
    if (d!=NULL) {
        RS_Graphic* graphic = (RS_Graphic*)d;
        if (graphic==NULL) {
            return;
        }

        RS_Ellipse* ellipse;
        RS_Line* line;

        for (double a=0.0; a<2*M_PI; a+=0.1) {
            RS_Vector v;
            v.setPolar(50.0, a);
            double xp = 1000.0*a;

            RS_EllipseData ellipseData(RS_Vector(xp,0.0),
                                       v,
                                       0.5,
                                       0.0, 2*M_PI,
                                       false);
            ellipse = new RS_Ellipse(graphic, ellipseData);

            ellipse->setPen(RS_Pen(RS_Color(255, 0, 255),
                                   RS2::Width01,
                                   RS2::SolidLine));

            graphic->addEntity(ellipse);
            //graphic->addEntity(new RS_Point(graphic, ellipse->getMax()));
            //graphic->addEntity(new RS_Point(graphic, ellipse->getMin()));

            line = new RS_Line(graphic,
                               RS_LineData(RS_Vector(xp,0.0),
                                           RS_Vector(xp,0.0)+v));
            line->setPen(RS_Pen(RS_Color(128, 128, 128),
                                RS2::Width01,
                                RS2::SolidLine));
            graphic->addEntity(line);


            /*
                     for (double mx=-60.0; mx<60.0; mx+=1.0) {
                         //for (double mx=0.0; mx<1.0; mx+=2.5) {
                         RS_VectorSolutions sol = ellipse->mapX(xp + mx);
                         //graphic->addEntity(new RS_Point(graphic,
                         //                   sol.vector2 + RS_Vector(a*500.0, 0.0)));
                         //graphic->addEntity(new RS_Point(graphic,
                         //                   sol.vector3 + RS_Vector(a*500.0, 0.0)));
                         //graphic->addEntity(new RS_Point(graphic,
                         //                   sol.vector4 + RS_Vector(a*500.0, 0.0)));

                         line = new RS_Line(graphic,
                                            RS_LineData(RS_Vector(xp+mx,-50.0),
                                                        RS_Vector(xp+mx,50.0)));
                         line->setPen(RS_Pen(RS_Color(60, 60, 60),
                                             RS2::Width01,
                                             RS2::SolidLine));
                         graphic->addEntity(line);

                         graphic->addEntity(new RS_Point(graphic,
                                                         sol.get(0)));
                     }
            */
        }


        // different minor/minor relations
        /*
              double x, y;
              for (y=-250.0; y<=250.0; y+=50.0) {
                  for (x=-250.0; x<=250.0; x+=50.0) {
                      RS_Vector v(x, y);

                      ellipse = new RS_Ellipse(graphic,
                                               v,
                                               RS_Vector((x/5+50.0)/2.0, 0.0), 
                                         fabs(x/y),
                                               0.0, 2*M_PI,
                                               false);

                ellipse->setPen(RS_Pen(RS_Color(255, 255, 0),
                                       RS2::Width01,
                                       RS2::DashDotLine));

                      graphic->addEntity(ellipse);
                      graphic->addEntity(new RS_Point(graphic, ellipse->getMax()));
                      graphic->addEntity(new RS_Point(graphic, ellipse->getMin()));

                ellipse = new RS_Ellipse(graphic,
                                               v + RS_Vector(750.0, 0.0),
                                               RS_Vector((x/5+50.0)/2.0, 0.0),
                                               fabs(x/y),
                                               2*M_PI, 0.0,
                                               true);

                      graphic->addEntity(ellipse);
                      graphic->addEntity(new RS_Point(graphic, ellipse->getMax()));
                      graphic->addEntity(new RS_Point(graphic, ellipse->getMin()));
                  }
              }
        */

        /*
              // different rotation angles:
              double rot;
              for (rot=0.0; rot<=2*M_PI+0.1; rot+=(M_PI/8)) {
                  ellipse = new RS_Ellipse(graphic,
                                           RS_Vector(rot*200, 500.0),
                                           RS_Vector(50.0, 0.0).rotate(rot),
                                           0.3,
                                           0.0, 2*M_PI,
                                           false);
                  graphic->addEntity(ellipse);
                  graphic->addEntity(new RS_Point(graphic, ellipse->getMax()));
                  graphic->addEntity(new RS_Point(graphic, ellipse->getMin()));
              }


              // different arc angles:
              double a1, a2;
              for (rot=0.0; rot<=2*M_PI+0.1; rot+=(M_PI/8)) {
                  for (a1=0.0; a1<=2*M_PI+0.1; a1+=(M_PI/8)) {
                      for (a2=a1+M_PI/8; a2<=2*M_PI+a1+0.1; a2+=(M_PI/8)) {
                          ellipse = new RS_Ellipse(graphic,
                                                   RS_Vector(-500.0-a1*200.0-5000.0*rot,
                                                             500.0-a2*200.0),
                                                   RS_Vector(50.0, 0.0).rotate(rot),
                                                   0.3,
                                                   a1, a2,
                                                   false);
                          graphic->addEntity(ellipse);
                          graphic->addEntity(new RS_Point(graphic, ellipse->getMax()));
                          graphic->addEntity(new RS_Point(graphic, ellipse->getMin()));
                      }
                  }
              }
        */

        RS_GraphicView* v = getGraphicView();
        if (v!=NULL) {
            v->redraw();
        }
    }
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestInsertText() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestInsertText()");


    RS_Document* d = getDocument();
    if (d!=NULL) {
        RS_Graphic* graphic = (RS_Graphic*)d;
        if (graphic==NULL) {
            return;
        }

        RS_Text* text;
        RS_TextData textData;

        textData = RS_TextData(RS_Vector(10.0,10.0),
                               10.0, 100.0,
                               RS2::VAlignTop,
                               RS2::HAlignLeft,
                               RS2::LeftToRight,
                               RS2::Exact,
                               1.0,
                               "Andrew",
                               "normal",
                               0.0);
        text = new RS_Text(graphic, textData);

        text->setLayerToActive();
        text->setPen(RS_Pen(RS_Color(255, 0, 0),
                            RS2::Width01,
                            RS2::SolidLine));
        graphic->addEntity(text);

        /*
              double x, y;
              for (y=-250.0; y<=250.0; y+=50.0) {
                  for (x=-250.0; x<=250.0; x+=50.0) {
                      RS_Vector v(x, y);

                      textData = RS_TextData(v,
                                             10.0, 100.0,
                                             RS2::VAlignTop,
                                             RS2::HAlignLeft,
                                             RS2::LeftToRight,
                                             RS2::Exact,
                                             1.0,
                                             "Andrew",
                                             "normal",
                                             0.0);

                      text = new RS_Text(graphic, textData);

                      text->setLayerToActive();
                      text->setPen(RS_Pen(RS_Color(255, 0, 0),
                                          RS2::Width01,
                                          RS2::SolidLine));
                      graphic->addEntity(text);
                  }
              }

              RS_Line* line;
              for (x=0.0; x<M_PI*2.0; x+=0.2) {
                  RS_Vector v(600.0+cos(x)*50.0, 0.0+sin(x)*50.0);

                  line = new RS_Line(graphic,
                                     RS_LineData(RS_Vector(600.0,0.0),
                                                 v));
                  line->setLayerToActive();
                  line->setPenToActive();
                  graphic->addEntity(line);

                  textData = RS_TextData(v,
                                         5.0, 50.0,
                                         RS2::VAlignTop,
                                         RS2::HAlignLeft,
                                         RS2::LeftToRight,
                                         RS2::Exact,
                                         1.0,
                                         "Andrew",
                                         "normal",
                                         x);

                  text = new RS_Text(graphic, textData);

                  text->setLayerToActive();
                  text->setPen(RS_Pen(RS_Color(255, 0, 0),
                                      RS2::Width01,
                                      RS2::SolidLine));
                  graphic->addEntity(text);
              }

              RS_SolidData solidData = RS_SolidData(RS_Vector(5.0, 10.0),
                                                    RS_Vector(25.0, 15.0),
                                                    RS_Vector(15.0, 30.0));

              RS_Solid* s = new RS_Solid(graphic, solidData);

              s->setLayerToActive();
              s->setPen(RS_Pen(RS_Color(255, 255, 0),
                               RS2::Width01,
                               RS2::SolidLine));
              graphic->addEntity(s);

              RS_GraphicView* v = getGraphicView();
              if (v!=NULL) {
                  v->redraw();
              }
        */
    }
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestInsertImage() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestInsertImage()");


    RS_Document* d = getDocument();
    if (d!=NULL) {
        RS_Graphic* graphic = (RS_Graphic*)d;
        if (graphic==NULL) {
            return;
        }

        RS_Image* image;
        RS_ImageData imageData;

        imageData = RS_ImageData(0, RS_Vector(50.0,30.0),
                                 RS_Vector(0.5,0.5),
                                 RS_Vector(-0.5,0.5),
                                 RS_Vector(640,480),
                                 "/home/andrew/data/image.png",
                                 50, 50, 0);
        image = new RS_Image(graphic, imageData);

        image->setLayerToActive();
        image->setPen(RS_Pen(RS_Color(255, 0, 0),
                             RS2::Width01,
                             RS2::SolidLine));
        graphic->addEntity(image);
    }
}

/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestMath01() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestMath01()");

    RS_Document* d = getDocument();
    if (d!=NULL) {
        RS_Graphic* graphic = (RS_Graphic*)d;
        if (graphic==NULL) {
            return;
        }

        // axis
        graphic->addEntity(new RS_Line(graphic,
                                       RS_LineData(RS_Vector(0.0,0.0),
                                                   RS_Vector(2*M_PI,0.0))));
        graphic->addEntity(new RS_Line(graphic,
                                       RS_LineData(RS_Vector(0.0,-1.0),
                                                   RS_Vector(0.0,1.0))));

        // cos
        double a;
        double x = 59.0/ARAD;
        double x_0 = 60.0/ARAD;
        for (a=0.01; a<2*M_PI; a+=0.01) {
            // cos curve:
            RS_Line* line = new RS_Line(graphic,
                                        RS_LineData(RS_Vector(a-0.01, cos(a-0.01)),
                                                    RS_Vector(a, cos(a))));
            graphic->addEntity(line);

            // tangent:
            graphic->addEntity(new RS_Line(graphic,
                                           RS_LineData(RS_Vector(a-0.01,cos(x_0)-sin(x_0)*(a-0.01-x_0)),
                                                       RS_Vector(a,cos(x_0)-sin(x_0)*(a-x_0)))));
        }

        // 59.0 deg
        graphic->addEntity(new RS_Line(graphic,
                                       RS_LineData(RS_Vector(x,0.0),
                                                   RS_Vector(x,1.0))));

        // 60.0 deg
        graphic->addEntity(new RS_Line(graphic,
                                       RS_LineData(RS_Vector(x_0,0.0),
                                                   RS_Vector(x_0,1.0))));

        // tangent
        //graphic->addEntity(new RS_Line(graphic,
        //                   RS_Vector(0.0,cos(x_0)-sin(x_0)*(0.0-x_0)),
        //                   RS_Vector(6.0,cos(x_0)-sin(x_0)*(6.0-x_0))));


        RS_GraphicView* v = getGraphicView();
        if (v!=NULL) {
            v->redraw();
        }
    }
}




/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestResize640() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestResize640()");

    resize(640, 480);
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestResize800() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestResize800()");

    resize(800, 600);
}



/**
 * Testing function.
 */
void QC_ApplicationWindow::slotTestResize1024() {
    RS_DEBUG->print("QC_ApplicationWindow::slotTestResize1024()");

    resize(1024, 768);
}

bool QC_ApplicationWindow::queryClose()
{
	QC_GraphicView *view;
	for (int i = 0; view = dynamic_cast<QC_GraphicView*>(tabWidget()
		->page(i)); i++) {
		if (view->getParent() && !view->getParent()->queryClose())
			return false;
	}
	return true;
}

/**
 * Handle hotkeys. Don't let it to the default handler of Qt.
 * it will consume them also if a text field is active
 * which means it's impossible to enter a command.
 */
void QC_ApplicationWindow::keyPressEvent(QKeyEvent* e) {

    // timer
    static QTime ts = QTime();
    static QString firstKey = "";

    // single key codes:
    switch (e->key()) {
    case Qt::Key_Shift:
    case Qt::Key_Control:
    case Qt::Key_Meta:
    case Qt::Key_Alt:
    case Qt::Key_CapsLock: {
            QMainWindow::keyPressEvent(e);

            // forward to actions:
            RS_GraphicView* graphicView = getGraphicView();
            if (graphicView!=NULL) {
                graphicView->keyPressEvent(e);
            }
            e->accept();
        }
        break;

    case Qt::Key_Escape:
        firstKey = "";
        slotBack();
        e->accept();
        break;

    case Qt::Key_Return:
        if (firstKey.isEmpty()) {
            slotEnter();
            e->accept();
        }
        break;

    case Qt::Key_Plus:
        if (firstKey.isEmpty()) {
            actionHandler->slotZoomIn();
            e->accept();
        }
        break;

    case Qt::Key_Minus:
        if (firstKey.isEmpty()) {
            actionHandler->slotZoomOut();
            e->accept();
        }
        break;

    default:
        e->ignore();
        break;
    }

    if (e->isAccepted()) {
        return;
    }

    QTime now = QTime::currentTime();

    // multi key codes:
    if (ts.msecsTo(now)<2000) {
        QString code =
            QString("%1%2").arg(firstKey).arg(QChar(e->key())).lower();

        if (actionHandler->keycode(code)==false) {
            ts = now;
            if (QChar(e->key()).isPrint()) {
                firstKey += e->key();
            }
        }
        else {
            firstKey="";
        }
    }
    else {
        ts = now;
        if (QChar(e->key()).isPrint()) {
            firstKey = e->key();
        }
    }

    QMainWindow::keyPressEvent(e);
}


void QC_ApplicationWindow::keyReleaseEvent(QKeyEvent* e) {

    switch (e->key()) {
    case Qt::Key_Shift:
    case Qt::Key_Control:
    case Qt::Key_Meta:
    case Qt::Key_Alt:
    case Qt::Key_CapsLock: {
            QMainWindow::keyReleaseEvent(e);

            // forward to actions:
            RS_GraphicView* graphicView = getGraphicView();
            if (graphicView!=NULL) {
                graphicView->keyReleaseEvent(e);
            }
            e->accept();
        }
        break;
    }

    QMainWindow::keyPressEvent(e);
}

#include "qc_applicationwindow.moc"

