/****************************************************************************
**
** Copyright (C) 2001-2003 RibbonSoft. All rights reserved.
** Copyright (c) 2006 Martin Teichmann. All rights reserved.
**
** This file is part of the KAD project, based on a file from
** the QCad project.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
**********************************************************************/

#include "qc_mdiwindow.h"

#include <kprinter.h>
#include <kfiledialog.h>

#include "rs_application.h"

#include "qg_exitdialog.h"

#include <qapplication.h>
#include <qcursor.h>

int QC_MDIWindow::idCounter = 0;

/**
 * Constructor.
 *
 * @param doc Pointer to an existing document of NULL if a new 
 *   document shall be created for this window.
 * @param parent Parent widget. Usually a workspace.
 */
QC_MDIWindow::QC_MDIWindow(RS_Document* doc, QWidget* parent)
        : KParts::ReadWritePart(parent, "KAD part")
{
    owner = false;
	forceClosing = false;
    initDoc(doc);
    graphicView = new QC_GraphicView(document, this);
    setWidget(graphicView);
    graphicView->setFocus();
    id = idCounter++;
    childWindows.setAutoDelete(false);
    parentWindow = NULL;

    if (document!=NULL) {
        if (document->getLayerList()!=NULL) {
            // Link the graphic view to the layer widget
            document->getLayerList()->addListener(graphicView);
        }
        if (document->getBlockList()!=NULL) {
            // Link the graphic view to the block widget
            document->getBlockList()->addListener(graphicView);
        }
    }
}



/**
 * Destructor.
 *
 * Deletes the document associated with this window.
 */
QC_MDIWindow::~QC_MDIWindow() {
	if (!document) return;

	if (owner) {
		QC_MDIWindow *child;
		for (child = childWindows.first(); child;
			child = childWindows.next()) {
			delete child;
		}
	}

	if (document->getLayerList()) {
		document->getLayerList()->removeListener(graphicView);
	}

	if (document->getBlockList()) {
		document->getBlockList()->removeListener(graphicView);
	}

        if (owner) delete document;
}



/**
 * Adds another MDI window to the list of known windows that
 * depend on this one. This can be another view or a view for
 * a particular block.
 */
void QC_MDIWindow::addChildWindow(QC_MDIWindow* w) {
    RS_DEBUG->print("RS_MDIWindow::addChildWindow()");

    childWindows.append(w);
    w->setParentWindow(this);

    RS_DEBUG->print("children: %d", childWindows.count());
}



/**
 * Removes a child window.
 *
 * @see addChildWindow
 */
void QC_MDIWindow::removeChildWindow(QC_MDIWindow* w)
{
    childWindows.remove(w);
}

/**
 * @return pointer to the print preview of this drawing or NULL.
 */
QC_MDIWindow* QC_MDIWindow::getPrintPreview() {
    for (uint i=0; i<childWindows.count(); ++i) {
        if (childWindows.at(i)->getGraphicView()->isPrintPreview()) {
			return childWindows.at(i);
		}
    }

    return 0;
}



/**
 * closes this MDI window.
 *
 * @param force Disable cancel button (demo versions)
 * @param ask Ask user before closing.
 */
bool QC_MDIWindow::close() {
    bool ret = false;

    bool isBlock = (parentWindow != 0);

    // This is a block and we don't need to ask the user for closing
    //   since it's still available in the parent drawing after closing.
    if (isBlock) {
        // tell parent window we're not here anymore.
        if (parentWindow) parentWindow->removeChildWindow(this);
	delete this;
        return true;
    }

    // This is a graphic document. ask user for closing.
    else if (queryClose()) {
        // close all child windows:
	while (childWindows.first()) childWindows.first()->close();
	delete this;
        return true;
    }

    // User decided not to close graphic document:
    else return false;
}

class Graphic: public RS_Graphic
{
    public:
	virtual void setModified(bool m);
	QC_MDIWindow *owner;

 	virtual ~Graphic() { }
};

void Graphic::setModified(bool m)
{
	owner->setModified(m);
	RS_Graphic::setModified(m);
}

/**
 * Init the document.
 *
 * @param type Document type. RS:EntityGraphic or RS2::EntityBlock
 * @param container Entity container to be used as document or NULL
 * if a new document should be created.
 */
void QC_MDIWindow::initDoc(RS_Document* doc)
{
    if (doc) {
        document = doc;
        owner = false;
    } else {
	Graphic *g = new Graphic();
	g->owner = this;
        document = g;
        document->newDoc();
        owner = true;
    }

}


/**
 * Called when the current pen (color, style, width) has changed.
 * Sets the active pen for the document in this MDI window.
 */
void QC_MDIWindow::slotPenChanged(RS_Pen pen) {
    RS_DEBUG->print("QC_MDIWindow::slotPenChanged() begin");
    if (document!=NULL) {
        document->setActivePen(pen);
    }
    RS_DEBUG->print("QC_MDIWindow::slotPenChanged() end");
}

void QC_MDIWindow::slotPenColorSelected(const QColor &c)
{
	if (document) {
		RS_Pen pen = document->getActivePen();
		pen.setColor(c);
		document->setActivePen(pen);
	}
}

/**
 * Creates a new empty document in this MDI window.
 */
void QC_MDIWindow::slotFileNew() {
    RS_DEBUG->print("QC_MDIWindow::slotFileNew begin");
    if (document!=NULL && graphicView!=NULL) {
        document->newDoc();
        graphicView->redraw();
    }
    RS_DEBUG->print("QC_MDIWindow::slotFileNew end");
}



/**
 * Opens the given file in this MDI window.
 */
bool QC_MDIWindow::openFile()
{
    	bool ret;

        document->newDoc();

		// cosmetics..
		RS_APP->processEvents(1000);
		
        ret = document->open(m_file, RS2::FormatUnknown);

        if (ret) graphicView->zoomAuto(false);
    	return ret;
}



/**
 * Saves the current file.
 *
 * @return true if the file was saved successfully.
 *         false if the file could not be saved or the document 
 *         is invalid.
 */
bool QC_MDIWindow::slotFileSave(bool &cancelled) {
    RS_DEBUG->print("QC_MDIWindow::slotFileSave()");
    bool ret = false;
    cancelled = false;

    if (document!=NULL) {
        if (document->getFilename().isEmpty()) {
            ret = slotFileSaveAs(cancelled);
        } else {
            QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );
            ret = document->save();
            QApplication::restoreOverrideCursor();
        }
    }

    return ret;
}



/**
 * Saves the current file. The user is asked for a new filename 
 * and format.
 *
 * @return true if the file was saved successfully or the user cancelled.
 *         false if the file could not be saved or the document 
 *         is invalid.
 */
bool QC_MDIWindow::slotFileSaveAs(bool &cancelled) {
    KURL url = KFileDialog::getSaveURL(QString::null, "*.dxf|DXF Files");
    if (url.isEmpty()) return true;
    return saveAs(url);
}

bool QC_MDIWindow::saveFile()
{
	return document->saveAs(m_file, RS2::FormatDXF);
}


/**
 * Requests the closing of this MDI window.
 *
 * @param force Force closing by disabling the cancel button (for demo versions).
 */
bool QC_MDIWindow::slotFileClose(bool force) {
    RS_DEBUG->print("QC_MDIWindow::slotFileClose()");

    bool succ = true;
    int exit = 0;

    if(document!=NULL && document->isModified()) {
        QG_ExitDialog dlg(widget());
        dlg.setForce(force);
        if (document->getFilename().isEmpty()) {
            dlg.setText(tr("Do you really want to close the drawing?"));
        } else {
			QString fn = document->getFilename();
			if (fn.length() > 50) {
				fn = QString("%1...%2").arg(fn.left(24)).arg(fn.right(24));
			}
            dlg.setText(tr("Do you really want to close the file\n%1?")
                        .arg(fn));
        }
        dlg.setTitle(tr("Closing Drawing"));

        bool again;
        bool cancelled;
        do {
            again = false;
            exit = dlg.exec();

            switch (exit) {
            case 0: // cancel
                succ = false;
                break;
            case 1: // leave
                succ = true;
                break;
            case 2: // save
                succ = slotFileSave(cancelled);
                again = !succ || cancelled;
                break;
            case 3: // save as
                succ = slotFileSaveAs(cancelled);
                again = !succ || cancelled;
                break;
            default:
                break;
            }
        } while (again);
    } else {
        succ = true;
    }

    return succ;
}



void QC_MDIWindow::slotFilePrint() {
    KPrinter printer;
    if (printer.setup(widget())) {
        QPainter painter;
        painter.begin(&printer);

        ///////////////////////////////////////////////////////////////////
        // TODO: Define printing by using the QPainter methods here

        painter.end();
    }
}



/**
 * Streams some info about an MDI window to stdout.
 */
std::ostream& operator << (std::ostream& os, QC_MDIWindow& w) {
    os << "QC_MDIWindow[" << w.getId() << "]:\n";
    if (w.parentWindow!=NULL) {
        os << "  parentWindow: " << w.parentWindow->getId() << "\n";
    } else {
        os << "  parentWindow: NULL\n";
    }

    for (uint i=0; i<w.childWindows.count(); ++i) {
        os << "  childWindow[" << i << "]: "
        << w.childWindows.at(i)->getId() << "\n";
    }
    return os;
}

#include "qc_mdiwindow.moc"

